# Copyright (C) 2020, 2024, Hitachi, Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
"""Common module for Hitachi HBSD Driver."""

from collections import defaultdict
import json
import re

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils

from cinder import coordination
from cinder import exception
from cinder.i18n import _
from cinder.volume import configuration
from cinder.volume.drivers.hitachi import hbsd_utils as utils
from cinder.volume import volume_types
from cinder.volume import volume_utils

_GROUP_NAME_MAX_LEN_FC = 64
_GROUP_NAME_MAX_LEN_ISCSI = 32

GROUP_NAME_ALLOWED_CHARS = 'a-zA-Z0-9.@_:-'
GROUP_NAME_VAR_WWN = '{wwn}'
GROUP_NAME_VAR_IP = '{ip}'
GROUP_NAME_VAR_HOST = '{host}'

_GROUP_NAME_VAR_WWN_LEN = 16
_GROUP_NAME_VAR_IP_LEN = 15
_GROUP_NAME_VAR_HOST_LEN = 1
_GROUP_NAME_VAR_LEN = {GROUP_NAME_VAR_WWN: _GROUP_NAME_VAR_WWN_LEN,
                       GROUP_NAME_VAR_IP: _GROUP_NAME_VAR_IP_LEN,
                       GROUP_NAME_VAR_HOST: _GROUP_NAME_VAR_HOST_LEN}

STR_VOLUME = 'volume'
STR_SNAPSHOT = 'snapshot'

_UUID_PATTERN = re.compile(r'^[\da-f]{32}$')

_INHERITED_VOLUME_OPTS = [
    'volume_backend_name',
    'volume_driver',
    'reserved_percentage',
    'use_multipath_for_image_xfer',
    'enforce_multipath_for_image_xfer',
    'max_over_subscription_ratio',
    'use_chap_auth',
    'chap_username',
    'chap_password',
]

COMMON_VOLUME_OPTS = [
    cfg.StrOpt(
        'hitachi_storage_id',
        default=None,
        help='Product number of the storage system.'),
    cfg.ListOpt(
        'hitachi_pools',
        default=[],
        deprecated_name='hitachi_pool',
        help='Pool number[s] or pool name[s] of the DP pool.'),
    cfg.StrOpt(
        'hitachi_snap_pool',
        default=None,
        help='Pool number or pool name of the snapshot pool.'),
    cfg.StrOpt(
        'hitachi_ldev_range',
        default=None,
        help='Range of the LDEV numbers in the format of \'xxxx-yyyy\' that '
             'can be used by the driver. Values can be in decimal format '
             '(e.g. 1000) or in colon-separated hexadecimal format '
             '(e.g. 00:03:E8).'),
    cfg.ListOpt(
        'hitachi_target_ports',
        default=[],
        help='IDs of the storage ports used to attach volumes to the '
             'controller node. To specify multiple ports, connect them by '
             'commas (e.g. CL1-A,CL2-A).'),
    cfg.ListOpt(
        'hitachi_compute_target_ports',
        default=[],
        help='IDs of the storage ports used to attach volumes to compute '
             'nodes. To specify multiple ports, connect them by commas '
             '(e.g. CL1-A,CL2-A).'),
    cfg.BoolOpt(
        'hitachi_group_create',
        default=False,
        help='If True, the driver will create host groups or iSCSI targets on '
             'storage ports as needed.'),
    cfg.BoolOpt(
        'hitachi_group_delete',
        default=False,
        help='If True, the driver will delete host groups or iSCSI targets on '
             'storage ports as needed.'),
    cfg.IntOpt(
        'hitachi_copy_speed',
        default=3,
        min=1, max=15,
        help='Copy speed of storage system. 1 or 2 indicates '
             'low speed, 3 indicates middle speed, and a value between 4 and '
             '15 indicates high speed.'),
    cfg.IntOpt(
        'hitachi_copy_check_interval',
        default=3,
        min=1, max=600,
        help='Interval in seconds to check copying status during a volume '
             'copy.'),
    cfg.IntOpt(
        'hitachi_async_copy_check_interval',
        default=10,
        min=1, max=600,
        help='Interval in seconds to check asynchronous copying status during '
             'a copy pair deletion or data restoration.'),
]

COMMON_PORT_OPTS = [
    cfg.BoolOpt(
        'hitachi_port_scheduler',
        default=False,
        help='Enable port scheduling of WWNs to the configured ports so that '
             'WWNs are registered to ports in a round-robin fashion.'),
]

COMMON_PAIR_OPTS = [
    cfg.IntOpt(
        'hitachi_pair_target_number',
        default=0, min=0, max=99,
        help='Pair target name of the host group or iSCSI target'),
]

COMMON_NAME_OPTS = [
    cfg.StrOpt(
        'hitachi_group_name_format',
        default=None,
        help='Format of host groups, iSCSI targets, and server objects.'),
]

CONF = cfg.CONF
CONF.register_opts(COMMON_VOLUME_OPTS, group=configuration.SHARED_CONF_GROUP)
CONF.register_opts(COMMON_PORT_OPTS, group=configuration.SHARED_CONF_GROUP)
CONF.register_opts(COMMON_PAIR_OPTS, group=configuration.SHARED_CONF_GROUP)
CONF.register_opts(COMMON_NAME_OPTS, group=configuration.SHARED_CONF_GROUP)

LOG = logging.getLogger(__name__)
MSG = utils.HBSDMsg


def str2int(num):
    """Convert a string into an integer."""
    if not num:
        return None
    if num.isdigit():
        return int(num)
    if not re.match(r'[0-9a-fA-F][0-9a-fA-F]:[0-9a-fA-F]' +
                    '[0-9a-fA-F]:[0-9a-fA-F][0-9a-fA-F]$', num):
        return None
    try:
        return int(num.replace(':', ''), 16)
    except ValueError:
        return None


class HBSDCommon():
    """Common class for Hitachi HBSD Driver."""

    def __init__(self, conf, driverinfo, db):
        """Initialize instance variables."""
        self.conf = conf
        self.db = db
        self.ctxt = None
        self.lock = {
            'do_setup': 'do_setup',
        }
        self.driver_info = driverinfo
        self.storage_info = {
            'protocol': driverinfo['proto'],
            'pool_id': None,
            'snap_pool_id': None,
            'ldev_range': [],
            'controller_ports': [],
            'compute_ports': [],
            'pair_ports': [],
            'wwns': {},
            'portals': {},
        }
        self.storage_id = None
        if self.storage_info['protocol'] == 'FC':
            self.group_name_format = {
                'group_name_max_len': _GROUP_NAME_MAX_LEN_FC,
                'group_name_var_cnt': {
                    GROUP_NAME_VAR_WWN: [1],
                    GROUP_NAME_VAR_IP: [0],
                    GROUP_NAME_VAR_HOST: [0, 1],
                },
                'group_name_format_default': self.driver_info[
                    'target_prefix'] + '{wwn}',
            }
        if self.storage_info['protocol'] == 'iSCSI':
            self.group_name_format = {
                'group_name_max_len': _GROUP_NAME_MAX_LEN_ISCSI,
                'group_name_var_cnt': {
                    GROUP_NAME_VAR_WWN: [0],
                    GROUP_NAME_VAR_IP: [1],
                    GROUP_NAME_VAR_HOST: [0, 1],
                },
                'group_name_format_default': self.driver_info[
                    'target_prefix'] + '{ip}',
            }
        self.format_info = {
            'group_name_format': self.group_name_format[
                'group_name_format_default'],
            'group_name_format_without_var_len': (
                len(re.sub('|'.join([GROUP_NAME_VAR_WWN,
                    GROUP_NAME_VAR_IP, GROUP_NAME_VAR_HOST]), '',
                    self.group_name_format['group_name_format_default']))),
            'group_name_var_cnt': {
                GROUP_NAME_VAR_WWN: self.group_name_format[
                    'group_name_format_default'].count(GROUP_NAME_VAR_WWN),
                GROUP_NAME_VAR_IP: self.group_name_format[
                    'group_name_format_default'].count(GROUP_NAME_VAR_IP),
                GROUP_NAME_VAR_HOST: self.group_name_format[
                    'group_name_format_default'].count(GROUP_NAME_VAR_HOST),
            }
        }

        self._required_common_opts = [
            self.driver_info['param_prefix'] + '_storage_id',
            self.driver_info['param_prefix'] + '_pools',
        ]
        self.port_index = {}

    def get_pool_id_of_volume(self, volume):
        pools = self._stats['pools']
        if len(pools) == 1:
            return pools[0]['location_info']['pool_id']
        pool_name = volume_utils.extract_host(volume['host'], 'pool')
        for pool in pools:
            if pool['pool_name'] == pool_name:
                return pool['location_info']['pool_id']
        return None

    def create_ldev(
            self, size, extra_specs, pool_id, ldev_range, qos_specs=None):
        """Create an LDEV and return its LDEV number."""
        raise NotImplementedError()

    def modify_ldev_name(self, ldev, name):
        """Modify LDEV name."""
        raise NotImplementedError()

    def create_volume(self, volume):
        """Create a volume and return its properties."""
        extra_specs = self.get_volume_extra_specs(volume)
        pool_id = self.get_pool_id_of_volume(volume)
        ldev_range = self.storage_info['ldev_range']
        qos_specs = utils.get_qos_specs_from_volume(volume)
        try:
            ldev = self.create_ldev(volume['size'], extra_specs, pool_id,
                                    ldev_range, qos_specs=qos_specs)
        except Exception:
            with excutils.save_and_reraise_exception():
                self.output_log(MSG.CREATE_LDEV_FAILED)
        self.modify_ldev_name(ldev, volume['id'].replace("-", ""))
        return {
            'provider_location': str(ldev),
        }

    def get_ldev_info(self, keys, ldev, **kwargs):
        """Return a dictionary of LDEV-related items."""
        raise NotImplementedError()

    def create_pair_on_storage(
            self, pvol, svol, snap_pool_id, is_snapshot=False):
        """Create a copy pair on the storage."""
        raise NotImplementedError()

    def wait_copy_completion(self, pvol, svol):
        """Wait until copy is completed."""
        raise NotImplementedError()

    def copy_on_storage(
            self, pvol, size, extra_specs, pool_id, snap_pool_id, ldev_range,
            is_snapshot=False, sync=False, is_rep=False, qos_specs=None):
        """Create a copy of the specified LDEV on the storage."""
        ldev_info = self.get_ldev_info(['status', 'attributes'], pvol)
        if ldev_info['status'] != 'NML':
            msg = self.output_log(MSG.INVALID_LDEV_STATUS_FOR_COPY, ldev=pvol)
            self.raise_error(msg)
        svol = self.create_ldev(
            size, extra_specs, pool_id, ldev_range, qos_specs=qos_specs)
        try:
            self.create_pair_on_storage(
                pvol, svol, snap_pool_id, is_snapshot=is_snapshot)
            if sync or is_rep:
                self.wait_copy_completion(pvol, svol)
        except Exception:
            with excutils.save_and_reraise_exception():
                try:
                    self.delete_ldev(svol)
                except exception.VolumeDriverException:
                    self.output_log(MSG.DELETE_LDEV_FAILED, ldev=svol)
        return svol

    def create_volume_from_src(self, volume, src, src_type, is_rep=False):
        """Create a volume from a volume or snapshot and return its properties.

        """
        ldev = self.get_ldev(src)
        if ldev is None:
            msg = self.output_log(
                MSG.INVALID_LDEV_FOR_VOLUME_COPY, type=src_type, id=src['id'])
            self.raise_error(msg)

        size = volume['size']
        extra_specs = self.get_volume_extra_specs(volume)
        pool_id = self.get_pool_id_of_volume(volume)
        snap_pool_id = self.storage_info['snap_pool_id']
        ldev_range = self.storage_info['ldev_range']
        qos_specs = utils.get_qos_specs_from_volume(volume)
        new_ldev = self.copy_on_storage(ldev, size, extra_specs, pool_id,
                                        snap_pool_id, ldev_range,
                                        is_rep=is_rep, qos_specs=qos_specs)
        self.modify_ldev_name(new_ldev, volume['id'].replace("-", ""))
        if is_rep:
            self.delete_pair(new_ldev)

        return {
            'provider_location': str(new_ldev),
        }

    def create_cloned_volume(self, volume, src_vref):
        """Create a clone of the specified volume and return its properties."""
        return self.create_volume_from_src(volume, src_vref, STR_VOLUME)

    def create_volume_from_snapshot(self, volume, snapshot):
        """Create a volume from a snapshot and return its properties."""
        return self.create_volume_from_src(volume, snapshot, STR_SNAPSHOT)

    def delete_pair_based_on_svol(self, pvol, svol_info):
        """Disconnect all volume pairs to which the specified S-VOL belongs."""
        raise NotImplementedError()

    def get_pair_info(self, ldev, ldev_info=None):
        """Return volume pair info(LDEV number, pair status and pair type)."""
        raise NotImplementedError()

    def delete_pair(self, ldev, ldev_info=None):
        """Disconnect all volume pairs to which the specified LDEV belongs.

        :param int ldev: The ID of the LDEV whose TI pair needs be deleted
        :param dict ldev_info: LDEV info
        :return: None
        :raises VolumeDriverException: if the LDEV is a P-VOL of a TI pair
        """
        pair_info = self.get_pair_info(ldev, ldev_info)
        if not pair_info:
            return
        if pair_info['pvol'] == ldev:
            self.output_log(
                MSG.UNABLE_TO_DELETE_PAIR, pvol=pair_info['pvol'])
            self.raise_busy()
        else:
            self.delete_pair_based_on_svol(
                pair_info['pvol'], pair_info['svol_info'][0])

    def find_all_mapped_targets_from_storage(self, targets, ldev):
        """Add all port-gids connected with the LDEV to the list."""
        raise NotImplementedError()

    def unmap_ldev(self, targets, ldev):
        """Delete the LUN between the specified LDEV and port-gid."""
        raise NotImplementedError()

    def unmap_ldev_from_storage(self, ldev):
        """Delete the connection between the specified LDEV and servers."""
        targets = {
            'list': [],
        }
        self.find_all_mapped_targets_from_storage(targets, ldev)
        self.unmap_ldev(targets, ldev)

    def delete_ldev_from_storage(self, ldev):
        """Delete the specified LDEV from the storage."""
        raise NotImplementedError()

    def delete_ldev(self, ldev, ldev_info=None):
        """Delete the specified LDEV.

        :param int ldev: The ID of the LDEV to be deleted
        :param dict ldev_info: LDEV info
        :return: None
        """
        self.delete_pair(ldev, ldev_info)
        self.unmap_ldev_from_storage(ldev)
        self.delete_ldev_from_storage(ldev)

    def is_invalid_ldev(self, ldev, obj, ldev_info_):
        """Check if the specified LDEV corresponds to the specified object.

        If the LDEV label and the object's id or name_id do not match, the LDEV
        was deleted and another LDEV with the same ID was created for another
        volume or snapshot. In this case, we say that the LDEV is invalid.
        If the LDEV label is not set or its format is unexpected, we cannot
        judge if the LDEV corresponds to the object. This can happen if the
        LDEV was created in older versions of this product or if the user
        overwrote the label. In this case, we just say that the LDEV is not
        invalid, although we are not completely sure about it.
        The reason for using name_id rather than id for volumes in comparison
        is that id of the volume that corresponds to the LDEV changes by
        host-assisted migration while that is not the case with name_id and
        that the LDEV label is created from id of the volume when the LDEV is
        created and is never changed after that.
        Because Snapshot objects do not have name_id, we use id instead of
        name_id if the object is a Snapshot. We assume that the object is a
        Snapshot object if hasattr(obj, 'name_id') returns False.
        This method returns False if the LDEV does not exist on the storage.
        The absence of the LDEV on the storage is detected elsewhere.
        :param int ldev: The ID of the LDEV to be checked
        :param obj: The object to be checked
        :type obj: Volume or Snapshot
        :param dict ldev_info_: LDEV info. This is an output area. Data is
        written by this method, but the area must be secured by the caller.
        :return: True if the LDEV does not correspond to the object, False
        otherwise
        :rtype: bool
        """
        ldev_info = self.get_ldev_info(None, ldev)
        # To avoid calling the same REST API multiple times, we pass the LDEV
        # info to the caller.
        ldev_info_.update(ldev_info)
        return ('label' in ldev_info
                and _UUID_PATTERN.match(ldev_info['label'])
                and ldev_info['label'] != (
                    obj.name_id if hasattr(obj, 'name_id') else
                    obj.id).replace('-', ''))

    def delete_volume(self, volume):
        """Delete the specified volume."""
        ldev = self.get_ldev(volume)
        if ldev is None:
            self.output_log(
                MSG.INVALID_LDEV_FOR_DELETION,
                method='delete_volume', id=volume['id'])
            return
        # Check if the LDEV corresponds to the volume.
        # To avoid KeyError when accessing a missing attribute, set the default
        # value to None.
        ldev_info = defaultdict(lambda: None)
        if self.is_invalid_ldev(ldev, volume, ldev_info):
            # If the LDEV is assigned to another object, skip deleting it.
            self.output_log(MSG.SKIP_DELETING_LDEV, obj='volume',
                            obj_id=volume.id, ldev=ldev,
                            ldev_label=ldev_info['label'])
            return
        try:
            self.delete_ldev(ldev, ldev_info)
        except exception.VolumeDriverException as ex:
            if utils.BUSY_MESSAGE in ex.msg:
                raise exception.VolumeIsBusy(volume_name=volume['name'])
            else:
                raise ex

    def create_snapshot(self, snapshot):
        """Create a snapshot from a volume and return its properties."""
        src_vref = snapshot.volume
        ldev = self.get_ldev(src_vref)
        if ldev is None:
            msg = self.output_log(
                MSG.INVALID_LDEV_FOR_VOLUME_COPY,
                type='volume', id=src_vref['id'])
            self.raise_error(msg)
        size = snapshot['volume_size']
        extra_specs = self.get_volume_extra_specs(snapshot['volume'])
        pool_id = self.get_pool_id_of_volume(snapshot['volume'])
        snap_pool_id = self.storage_info['snap_pool_id']
        ldev_range = self.storage_info['ldev_range']
        qos_specs = utils.get_qos_specs_from_volume(snapshot)
        new_ldev = self.copy_on_storage(
            ldev, size, extra_specs, pool_id, snap_pool_id, ldev_range,
            is_snapshot=True, qos_specs=qos_specs)
        self.modify_ldev_name(new_ldev, snapshot.id.replace("-", ""))
        return {
            'provider_location': str(new_ldev),
        }

    def delete_snapshot(self, snapshot):
        """Delete the specified snapshot."""
        ldev = self.get_ldev(snapshot)
        if ldev is None:
            self.output_log(
                MSG.INVALID_LDEV_FOR_DELETION, method='delete_snapshot',
                id=snapshot['id'])
            return
        # Check if the LDEV corresponds to the snapshot.
        # To avoid KeyError when accessing a missing attribute, set the default
        # value to None.
        ldev_info = defaultdict(lambda: None)
        if self.is_invalid_ldev(ldev, snapshot, ldev_info):
            # If the LDEV is assigned to another object, skip deleting it.
            self.output_log(MSG.SKIP_DELETING_LDEV, obj='snapshot',
                            obj_id=snapshot.id, ldev=ldev,
                            ldev_label=ldev_info['label'])
            return
        try:
            self.delete_ldev(ldev, ldev_info)
        except exception.VolumeDriverException as ex:
            if utils.BUSY_MESSAGE in ex.msg:
                raise exception.SnapshotIsBusy(snapshot_name=snapshot['name'])
            else:
                raise ex

    def get_pool_info(self, pool_id, result=None):
        """Return the total and free capacity of the storage pool."""
        raise NotImplementedError()

    def get_pool_infos(self, pool_ids):
        """Return the total and free capacity of the storage pools."""
        raise NotImplementedError()

    def _create_single_pool_data(self, pool_id, pool_name, cap_data):
        location_info = {
            'storage_id': self.conf.hitachi_storage_id,
            'pool_id': pool_id,
            'snap_pool_id': self.storage_info['snap_pool_id'],
            'ldev_range': self.storage_info['ldev_range']}
        single_pool = {}
        single_pool.update(dict(
            pool_name=pool_name,
            reserved_percentage=self.conf.safe_get('reserved_percentage'),
            QoS_support=True,
            thin_provisioning_support=True,
            thick_provisioning_support=False,
            multiattach=True,
            consistencygroup_support=True,
            consistent_group_snapshot_enabled=True,
            max_over_subscription_ratio=(
                volume_utils.get_max_over_subscription_ratio(
                    self.conf.safe_get('max_over_subscription_ratio'),
                    True)),
            location_info=location_info
        ))
        if cap_data is None:
            single_pool.update(dict(
                total_capacity_gb=0,
                free_capacity_gb=0,
                provisioned_capacity_gb=0,
                backend_state='down'))
            self.output_log(MSG.POOL_INFO_RETRIEVAL_FAILED, pool=pool_name)
            return single_pool
        total_capacity, free_capacity, provisioned_capacity = cap_data
        single_pool.update(dict(
            total_capacity_gb=total_capacity,
            free_capacity_gb=free_capacity,
            provisioned_capacity_gb=provisioned_capacity
        ))
        single_pool.update(dict(backend_state='up'))
        return single_pool

    def update_volume_stats(self):
        """Update properties, capabilities and current states of the driver."""
        data = {}
        backend_name = (self.conf.safe_get('volume_backend_name') or
                        self.driver_info['volume_backend_name'])
        data = {
            'volume_backend_name': backend_name,
            'vendor_name': self.driver_info['vendor_name'],
            'driver_version': self.driver_info['version'],
            'storage_protocol': self.storage_info['protocol'],
            'pools': [],
        }
        for pool_id, pool_name, cap_data in zip(
                self.storage_info['pool_id'], self.conf.hitachi_pools,
                self.get_pool_infos(self.storage_info['pool_id'])):
            single_pool = self._create_single_pool_data(
                pool_id, pool_name if len(self.conf.hitachi_pools) > 1 else
                data['volume_backend_name'], cap_data)
            data['pools'].append(single_pool)
        LOG.debug("Updating volume status. (%s)", data)
        self._stats = data
        return data

    def discard_zero_page(self, volume):
        """Return the volume's no-data pages to the storage pool."""
        raise NotImplementedError()

    def check_pair_svol(self, ldev):
        """Check if the specified LDEV is S-VOL in a copy pair."""
        raise NotImplementedError()

    def extend_ldev(self, ldev, old_size, new_size):
        """Extend the specified LDEV to the specified new size."""
        raise NotImplementedError()

    def extend_volume(self, volume, new_size):
        """Extend the specified volume to the specified size."""
        ldev = self.get_ldev(volume)
        if ldev is None:
            msg = self.output_log(MSG.INVALID_LDEV_FOR_EXTENSION,
                                  volume_id=volume['id'])
            self.raise_error(msg)
        if self.check_pair_svol(ldev):
            msg = self.output_log(MSG.INVALID_VOLUME_TYPE_FOR_EXTEND,
                                  volume_id=volume['id'])
            self.raise_error(msg)
        self.delete_pair(ldev)
        self.extend_ldev(ldev, volume['size'], new_size)

    def get_ldev_by_name(self, name):
        """Get the LDEV number from the given name."""
        raise NotImplementedError()

    def check_ldev_manageability(self, ldev, existing_ref):
        """Check if the LDEV meets the criteria for being managed."""
        raise NotImplementedError()

    def get_qos_specs_from_ldev(self, ldev):
        raise NotImplementedError()

    def change_qos_specs(self, ldev, old_qos_specs, new_qos_specs):
        raise NotImplementedError()

    def manage_existing(self, volume, existing_ref):
        """Return volume properties which Cinder needs to manage the volume."""
        if 'source-name' in existing_ref:
            ldev = self.get_ldev_by_name(
                existing_ref.get('source-name').replace('-', ''))
        elif 'source-id' in existing_ref:
            ldev = str2int(existing_ref.get('source-id'))
        self.check_ldev_manageability(ldev, existing_ref)
        self.modify_ldev_name(ldev, volume['id'].replace("-", ""))
        new_qos_specs = utils.get_qos_specs_from_volume(volume)
        old_qos_specs = self.get_qos_specs_from_ldev(ldev)
        if old_qos_specs != new_qos_specs:
            self.change_qos_specs(ldev, old_qos_specs, new_qos_specs)
        return {
            'provider_location': str(ldev),
        }

    def get_ldev_size_in_gigabyte(self, ldev, existing_ref):
        """Return the size[GB] of the specified LDEV."""
        raise NotImplementedError()

    def manage_existing_get_size(self, volume, existing_ref):
        """Return the size[GB] of the specified volume."""
        ldev = None
        if 'source-name' in existing_ref:
            ldev = self.get_ldev_by_name(
                existing_ref.get('source-name').replace("-", ""))
        elif 'source-id' in existing_ref:
            ldev = str2int(existing_ref.get('source-id'))
        if ldev is None:
            msg = self.output_log(MSG.INVALID_LDEV_FOR_MANAGE)
            raise exception.ManageExistingInvalidReference(
                existing_ref=existing_ref, reason=msg)
        return self.get_ldev_size_in_gigabyte(ldev, existing_ref)

    def unmanage(self, volume):
        """Prepare the volume for removing it from Cinder management."""
        ldev = self.get_ldev(volume)
        if ldev is None:
            self.output_log(MSG.INVALID_LDEV_FOR_DELETION, method='unmanage',
                            id=volume['id'])
            return
        if self.check_pair_svol(ldev):
            self.output_log(
                MSG.INVALID_LDEV_TYPE_FOR_UNMANAGE, volume_id=volume['id'],
                volume_type=utils.NORMAL_LDEV_TYPE)
            raise exception.VolumeIsBusy(volume_name=volume['name'])
        try:
            self.delete_pair(ldev)
        except exception.VolumeDriverException as ex:
            if utils.BUSY_MESSAGE in ex.msg:
                raise exception.VolumeIsBusy(volume_name=volume['name'])
            else:
                raise ex

    def _range2list(self, param):
        """Analyze a 'xxx-xxx' string and return a list of two integers."""
        values = [str2int(value) for value in
                  self.conf.safe_get(param).split('-')]
        if len(values) != 2 or None in values or values[0] > values[1]:
            msg = self.output_log(MSG.INVALID_PARAMETER, param=param)
            self.raise_error(msg)
        return values

    def check_param_fc(self):
        """Check FC-related parameter values and consistency among them."""
        if hasattr(
                self.conf,
                self.driver_info['param_prefix'] + '_port_scheduler'):
            self.check_opts(self.conf, COMMON_PORT_OPTS)
            if (self.conf.hitachi_port_scheduler and
                    not self.conf.hitachi_group_create):
                msg = self.output_log(
                    MSG.INVALID_PARAMETER,
                    param=self.driver_info['param_prefix'] + '_port_scheduler')
                self.raise_error(msg)
            if (self._lookup_service is None and
                    self.conf.hitachi_port_scheduler):
                msg = self.output_log(MSG.ZONE_MANAGER_IS_NOT_AVAILABLE)
                self.raise_error(msg)

    def check_param_iscsi(self):
        """Check iSCSI-related parameter values and consistency among them."""
        if self.conf.use_chap_auth:
            if not self.conf.chap_username:
                msg = self.output_log(MSG.INVALID_PARAMETER,
                                      param='chap_username')
                self.raise_error(msg)
            if not self.conf.chap_password:
                msg = self.output_log(MSG.INVALID_PARAMETER,
                                      param='chap_password')
                self.raise_error(msg)

    def check_param(self):
        """Check parameter values and consistency among them."""
        self.check_opt_value(self.conf, _INHERITED_VOLUME_OPTS)
        self.check_opts(self.conf, COMMON_VOLUME_OPTS)
        if hasattr(
                self.conf,
                self.driver_info['param_prefix'] + '_pair_target_number'):
            self.check_opts(self.conf, COMMON_PAIR_OPTS)
        if hasattr(
                self.conf,
                self.driver_info['param_prefix'] + '_group_name_format'):
            self.check_opts(self.conf, COMMON_NAME_OPTS)
        if self.conf.hitachi_ldev_range:
            self.storage_info['ldev_range'] = self._range2list(
                self.driver_info['param_prefix'] + '_ldev_range')
        if (not self.conf.hitachi_target_ports and
                not self.conf.hitachi_compute_target_ports):
            msg = self.output_log(
                MSG.INVALID_PARAMETER,
                param=self.driver_info['param_prefix'] + '_target_ports or ' +
                self.driver_info['param_prefix'] + '_compute_target_ports')
            self.raise_error(msg)
        self._check_param_group_name_format()
        if (self.conf.hitachi_group_delete and
                not self.conf.hitachi_group_create):
            msg = self.output_log(
                MSG.INVALID_PARAMETER,
                param=self.driver_info['param_prefix'] + '_group_delete or '
                + self.driver_info['param_prefix'] + '_group_create')
            self.raise_error(msg)
        for opt in self._required_common_opts:
            if not self.conf.safe_get(opt):
                msg = self.output_log(MSG.INVALID_PARAMETER, param=opt)
                self.raise_error(msg)
        for pool in self.conf.hitachi_pools:
            if len(pool) == 0:
                msg = self.output_log(
                    MSG.INVALID_PARAMETER,
                    param=self.driver_info['param_prefix'] + '_pools')
                self.raise_error(msg)
        if self.storage_info['protocol'] == 'FC':
            self.check_param_fc()
        if self.storage_info['protocol'] == 'iSCSI':
            self.check_param_iscsi()

    def _check_param_group_name_format(self):
        if not hasattr(
                self.conf,
                self.driver_info['param_prefix'] + '_group_name_format'):
            return
        if self.conf.hitachi_group_name_format is not None:
            error_flag = False
            if re.match(
                    self.driver_info['target_prefix'] + '(' +
                    GROUP_NAME_VAR_WWN + '|' +
                    GROUP_NAME_VAR_IP + '|' + GROUP_NAME_VAR_HOST + '|' + '[' +
                    GROUP_NAME_ALLOWED_CHARS + '])+$',
                    self.conf.hitachi_group_name_format) is None:
                error_flag = True
            if not error_flag:
                for var in _GROUP_NAME_VAR_LEN:
                    self.format_info['group_name_var_cnt'][var] = (
                        self.conf.hitachi_group_name_format.count(var))
                    if (self.format_info[
                            'group_name_var_cnt'][var] not in
                            self.group_name_format['group_name_var_cnt'][var]):
                        error_flag = True
                        break
            if not error_flag:
                group_name_var_replaced = self.conf.hitachi_group_name_format
                for var, length in _GROUP_NAME_VAR_LEN.items():
                    group_name_var_replaced = (
                        group_name_var_replaced.replace(var, '_' * length))
                if len(group_name_var_replaced) > self.group_name_format[
                        'group_name_max_len']:
                    error_flag = True
            if error_flag:
                msg = self.output_log(
                    MSG.INVALID_PARAMETER,
                    param=self.driver_info['param_prefix'] +
                    '_group_name_format')
                self.raise_error(msg)
            self.format_info['group_name_format'] = (
                self.conf.hitachi_group_name_format)
            self.format_info['group_name_format_without_var_len'] = (
                len(re.sub('|'.join(
                    [GROUP_NAME_VAR_WWN, GROUP_NAME_VAR_IP,
                     GROUP_NAME_VAR_HOST]), '',
                    self.format_info['group_name_format'])))

    def need_client_setup(self):
        """Check if the making of the communication client is necessary."""
        raise NotImplementedError()

    def setup_client(self):
        """Initialize RestApiClient."""
        pass

    def enter_keep_session(self):
        """Begin the keeping of the session."""
        pass

    def check_pool_id(self):
        """Check the pool id of hitachi_pools and hitachi_snap_pool."""
        raise NotImplementedError()

    def connect_storage(self):
        """Prepare for using the storage."""
        self.check_pool_id()
        self.output_log(MSG.SET_CONFIG_VALUE, object='DP Pool ID',
                        value=self.storage_info['pool_id'])
        self.storage_info['controller_ports'] = []
        self.storage_info['compute_ports'] = []

    def find_targets_from_storage(self, targets, connector, target_ports):
        """Find mapped ports, memorize them and return unmapped port count."""
        raise NotImplementedError()

    def get_hba_ids_from_connector(self, connector):
        """Return the HBA ID stored in the connector."""
        if self.driver_info['hba_id'] in connector:
            return connector[self.driver_info['hba_id']]
        msg = self.output_log(MSG.RESOURCE_NOT_FOUND,
                              resource=self.driver_info['hba_id_type'])
        self.raise_error(msg)

    def set_device_map(self, targets, hba_ids, volume):
        return None, hba_ids

    def get_port_scheduler_param(self):
        if hasattr(
                self.conf,
                self.driver_info['param_prefix'] + '_port_scheduler'):
            return self.conf.hitachi_port_scheduler
        else:
            return False

    def create_target_by_port_scheduler(
            self, devmap, targets, connector, volume):
        raise NotImplementedError()

    def create_target_to_storage(self, port, connector, hba_ids):
        """Create a host group or an iSCSI target on the specified port."""
        raise NotImplementedError()

    def get_gid_from_targets(self, targets, port):
        for target_port, target_gid in targets['list']:
            if target_port == port:
                return target_gid
        msg = self.output_log(MSG.NO_CONNECTED_TARGET)
        self.raise_error(msg)

    def set_target_mode(self, port, gid):
        """Configure the target to meet the environment."""
        raise NotImplementedError()

    def set_hba_ids(self, port, gid, hba_ids):
        """Connect all specified HBAs with the specified port."""
        raise NotImplementedError()

    def delete_target_from_storage(self, port, gid):
        """Delete the host group or the iSCSI target from the port."""
        raise NotImplementedError()

    def set_target_map_info(self, targets, hba_ids, port):
        pass

    def create_target(self, targets, port, connector, hba_ids):
        """Create a host group or an iSCSI target on the storage port."""
        if port not in targets['info'] or not targets['info'][port]:
            target_name, gid = self.create_target_to_storage(
                port, connector, hba_ids)
            self.output_log(
                MSG.OBJECT_CREATED,
                object='a target',
                details='port: %(port)s, gid: %(gid)s, target_name: '
                        '%(target)s' %
                        {'port': port, 'gid': gid, 'target': target_name})
        else:
            gid = self.get_gid_from_targets(targets, port)
        try:
            if port not in targets['info'] or not targets['info'][port]:
                self.set_target_mode(port, gid)
            self.set_hba_ids(port, gid, hba_ids)
        except Exception:
            with excutils.save_and_reraise_exception():
                self.delete_target_from_storage(port, gid)
        targets['info'][port] = True
        if (port, gid) not in targets['list']:
            targets['list'].append((port, gid))
        self.set_target_map_info(targets, hba_ids, port)

    def create_mapping_targets(self, targets, connector, volume=None):
        """Create server-storage connection for all specified storage ports."""
        active_hba_ids = []
        hba_ids = self.get_hba_ids_from_connector(connector)

        devmap, active_hba_ids = self.set_device_map(targets, hba_ids, volume)

        if self.get_port_scheduler_param():
            self.create_target_by_port_scheduler(
                devmap, targets, connector, volume)
        else:
            for port in targets['info'].keys():
                if targets['info'][port]:
                    continue

                try:
                    self.create_target(
                        targets, port, connector, active_hba_ids)
                except exception.VolumeDriverException:
                    self.output_log(
                        self.driver_info['msg_id']['target'], port=port)

        # When other threads created a host group at same time, need to
        # re-find targets.
        if not targets['list']:
            self.find_targets_from_storage(
                targets, connector, list(targets['info'].keys()))

    def get_port_index_to_be_used(self, ports, network_name):
        backend_name = self.conf.safe_get('volume_backend_name')
        code = (
            str(self.conf.hitachi_storage_id) + backend_name + network_name)
        if code in self.port_index.keys():
            if self.port_index[code] >= len(ports) - 1:
                self.port_index[code] = 0
            else:
                self.port_index[code] += 1
        else:
            self.port_index[code] = 0

        return self.port_index[code]

    def init_cinder_hosts(self, **kwargs):
        """Initialize server-storage connection."""
        targets = kwargs.pop(
            'targets', {'info': {}, 'list': [], 'iqns': {}, 'target_map': {}})
        connector = volume_utils.brick_get_connector_properties(
            multipath=self.conf.use_multipath_for_image_xfer,
            enforce_multipath=self.conf.enforce_multipath_for_image_xfer)
        target_ports = self.storage_info['controller_ports']

        if target_ports:
            if (self.find_targets_from_storage(
                    targets, connector, target_ports) and
                    self.conf.hitachi_group_create):
                self.create_mapping_targets(targets, connector)

            self.require_target_existed(targets)

    def do_setup(self, context):
        """Prepare for the startup of the driver."""

        @coordination.synchronized('{self.lock[do_setup]}')
        def _with_synchronized(self):
            self.connect_storage()
            self.init_cinder_hosts()

        self.ctxt = context
        self.check_param()
        if self.need_client_setup():
            self.setup_client()
            self.enter_keep_session()
        _with_synchronized(self)

    def check_ports_info(self):
        """Check if available storage ports exist."""
        if (self.conf.hitachi_target_ports and
                not self.storage_info['controller_ports']):
            msg = self.output_log(MSG.RESOURCE_NOT_FOUND,
                                  resource="Target ports")
            self.raise_error(msg)
        if (self.conf.hitachi_compute_target_ports and
                not self.storage_info['compute_ports']):
            msg = self.output_log(MSG.RESOURCE_NOT_FOUND,
                                  resource="Compute target ports")
            self.raise_error(msg)
        self.output_log(MSG.SET_CONFIG_VALUE, object='target port list',
                        value=self.storage_info['controller_ports'])
        self.output_log(MSG.SET_CONFIG_VALUE,
                        object='compute target port list',
                        value=self.storage_info['compute_ports'])

    def attach_ldev(
            self, volume, ldev, connector, is_snapshot, targets, lun=None):
        """Initialize connection between the server and the volume."""
        raise NotImplementedError()

    def get_properties_fc(self, targets):
        """Return FC-specific server-LDEV connection info."""
        data = {}
        data['target_wwn'] = [
            self.storage_info['wwns'][target[0]] for target in targets['list']
            if targets['lun'][target[0]]]
        return data

    def get_properties_iscsi(self, targets, multipath):
        """Return iSCSI-specific server-LDEV connection info."""
        data = {}
        primary_target = targets['list'][0]
        if not multipath:
            data['target_portal'] = self.storage_info[
                'portals'][primary_target[0]]
            data['target_iqn'] = targets['iqns'][primary_target]
        else:
            # Set the list of numbers that LUN was added
            data['target_portals'] = [
                self.storage_info['portals'][target[0]] for target in
                targets['list'] if targets['lun'][target[0]]]
            data['target_iqns'] = [
                targets['iqns'][target] for target in targets['list']
                if targets['lun'][target[0]]]
        if self.conf.use_chap_auth:
            data['auth_method'] = 'CHAP'
            data['auth_username'] = self.conf.chap_username
            data['auth_password'] = self.conf.chap_password
        return data

    def get_properties(self, targets, target_lun, connector):
        """Return server-LDEV connection info."""
        multipath = connector.get('multipath', False)
        if self.storage_info['protocol'] == 'FC':
            data = self.get_properties_fc(targets)
        elif self.storage_info['protocol'] == 'iSCSI':
            data = self.get_properties_iscsi(targets, multipath)
        data['target_discovered'] = False
        if not multipath or self.storage_info['protocol'] == 'FC':
            data['target_lun'] = target_lun
        else:
            # Set the list of numbers that LUN was added
            target_luns = []
            for target in targets['list']:
                if targets['lun'][target[0]]:
                    target_luns.append(target_lun)
            data['target_luns'] = target_luns
        return data

    # A synchronization to prevent conflicts between host group creation
    # and deletion.
    @coordination.synchronized(
        '{self.driver_info[driver_file_prefix]}-host-'
        '{self.conf.hitachi_storage_id}-{connector[host]}')
    def initialize_connection(
            self, volume, connector, is_snapshot=False, lun=None):
        """Initialize connection between the server and the volume."""
        targets = {
            'info': {},
            'list': [],
            'lun': {},
            'iqns': {},
            'target_map': {},
        }
        ldev = self.get_ldev(volume)
        if ldev is None:
            msg = self.output_log(MSG.INVALID_LDEV_FOR_CONNECTION,
                                  volume_id=volume['id'])
            self.raise_error(msg)

        target_lun = self.attach_ldev(
            volume, ldev, connector, is_snapshot, targets, lun)

        return {
            'driver_volume_type': self.driver_info['volume_type'],
            'data': self.get_properties(targets, target_lun, connector),
        }, targets['target_map']

    def get_target_ports(self, connector):
        """Return a list of ports corresponding to the specified connector."""
        if 'ip' in connector and connector['ip'] == CONF.my_ip:
            return self.storage_info['controller_ports']
        return (self.storage_info['compute_ports'] or
                self.storage_info['controller_ports'])

    def get_port_hostgroup_map(self, ldev_id):
        """Get the mapping of a port and host group."""
        raise NotImplementedError()

    def set_terminate_target(self, fake_connector, port_hostgroup_map):
        """Set necessary information in connector in terminate."""
        raise NotImplementedError()

    def detach_ldev(self, volume, ldev, connector):
        """Terminate connection between the server and the volume."""
        raise NotImplementedError()

    def terminate_connection(self, volume, connector):
        """Terminate connection between the server and the volume."""
        ldev = self.get_ldev(volume)
        if ldev is None:
            self.output_log(MSG.INVALID_LDEV_FOR_UNMAPPING,
                            volume_id=volume['id'])
            return
        # If a fake connector is generated by nova when the host
        # is down, then the connector will not have a host property,
        # In this case construct the lock without the host property
        # so that all the fake connectors to an SVC are serialized
        if 'host' not in connector:
            port_hostgroup_map = self.get_port_hostgroup_map(ldev)
            if not port_hostgroup_map:
                self.output_log(MSG.NO_LUN, ldev=ldev)
                return
            self.set_terminate_target(connector, port_hostgroup_map)

        # A synchronization to prevent conflicts between host group creation
        # and deletion.
        @coordination.synchronized(
            '%(prefix)s-host-%(storage_id)s-%(host)s' % {
                'prefix': self.driver_info['driver_file_prefix'],
                'storage_id': self.conf.hitachi_storage_id,
                'host': connector.get('host'),
            }
        )
        def inner(self, volume, connector):
            deleted_targets = self.detach_ldev(volume, ldev, connector)
            if self.storage_info['protocol'] == 'FC':
                target_wwn = [
                    self.storage_info['wwns'][target]
                    for target in deleted_targets]
                return {'driver_volume_type': self.driver_info['volume_type'],
                        'data': {'target_wwn': target_wwn}}
        return inner(self, volume, connector)

    def filter_target_ports(self, target_ports, volume, is_snapshot=False):
        specs = self.get_volume_extra_specs(volume) if volume else None
        if not specs:
            return target_ports
        if self.driver_info.get('driver_dir_name'):
            if getattr(self, 'is_secondary', False):
                tps_name = self.driver_info[
                    'driver_dir_name'] + ':remote_target_ports'
            else:
                tps_name = self.driver_info[
                    'driver_dir_name'] + ':target_ports'
        else:
            return target_ports

        tps = specs.get(tps_name)
        if tps is None:
            return target_ports

        tpsset = set([s.strip() for s in tps.split(',')])
        filtered_tps = list(tpsset.intersection(target_ports))
        if is_snapshot:
            volume = volume['volume']
        for port in tpsset:
            if port not in target_ports:
                self.output_log(
                    MSG.INVALID_EXTRA_SPEC_KEY_PORT,
                    port=port, target_ports_param=tps_name,
                    volume_type=volume['volume_type']['name'])

        return filtered_tps

    def clean_mapping_targets(self, targets):
        raise NotImplementedError()

    def unmanage_snapshot(self, snapshot):
        """Output error message and raise NotImplementedError."""
        self.output_log(
            MSG.SNAPSHOT_UNMANAGE_FAILED, snapshot_id=snapshot['id'])
        raise NotImplementedError()

    def migrate_volume(self, volume, host):
        """Migrate the specified volume."""
        return False

    def update_migrated_volume(self, new_volume):
        """Return model update for migrated volume."""
        return {'_name_id': new_volume.name_id,
                'provider_location': new_volume.provider_location}

    def retype(self, ctxt, volume, new_type, diff, host):
        """Retype the specified volume."""
        return False

    def has_snap_pair(self, pvol, svol):
        """Check if the volume have the pair of the snapshot."""
        raise NotImplementedError()

    def restore_ldev(self, pvol, svol):
        """Restore a pair of the specified LDEV."""
        raise NotImplementedError()

    def revert_to_snapshot(self, volume, snapshot):
        """Rollback the specified snapshot."""
        pvol = self.get_ldev(volume)
        svol = self.get_ldev(snapshot)
        if (pvol is not None and
                svol is not None and
                self.has_snap_pair(pvol, svol)):
            self.restore_ldev(pvol, svol)
        else:
            raise NotImplementedError()

    def create_group(self):
        raise NotImplementedError()

    def delete_group(self, group, volumes):
        raise NotImplementedError()

    def create_group_from_src(
            self, context, group, volumes, snapshots=None, source_vols=None):
        raise NotImplementedError()

    def update_group(self, group, add_volumes=None):
        raise NotImplementedError()

    def create_group_snapshot(self, context, group_snapshot, snapshots):
        raise NotImplementedError()

    def delete_group_snapshot(self, group_snapshot, snapshots):
        raise NotImplementedError()

    def output_log(self, msg_enum, **kwargs):
        if self.storage_id is not None:
            return utils.output_log(
                msg_enum, storage_id=self.storage_id, **kwargs)
        else:
            return utils.output_log(msg_enum, **kwargs)

    def get_ldev(self, obj, both=False):
        if not obj:
            return None
        provider_location = obj.get('provider_location')
        if not provider_location:
            return None
        if provider_location.isdigit() and not getattr(self, 'is_secondary',
                                                       False):
            # This format implies that the value is the ID of an LDEV in the
            # primary storage. Therefore, the secondary instance should not
            # retrieve this value.
            return int(provider_location)
        if provider_location.startswith('{'):
            loc = json.loads(provider_location)
            if isinstance(loc, dict):
                if getattr(self, 'is_primary', False) or (
                        hasattr(self, 'primary_storage_id') and not both):
                    return None if 'pldev' not in loc else int(loc['pldev'])
                elif getattr(self, 'is_secondary', False):
                    return None if 'sldev' not in loc else int(loc['sldev'])
                if hasattr(self, 'primary_storage_id'):
                    return {key: loc.get(key) for key in ['pldev', 'sldev']}
        return None

    def check_opt_value(self, conf, names):
        """Check if the parameter names and values are valid."""
        for name in names:
            try:
                getattr(conf, name)
            except (cfg.NoSuchOptError, cfg.ConfigFileValueError):
                with excutils.save_and_reraise_exception():
                    self.output_log(MSG.INVALID_PARAMETER, param=name)

    def check_opts(self, conf, opts):
        """Check if the specified configuration is valid."""
        names = []
        for opt in opts:
            if opt.required and not conf.safe_get(opt.name):
                msg = self.output_log(MSG.INVALID_PARAMETER, param=opt.name)
                self.raise_error(msg)
            names.append(opt.name)
        self.check_opt_value(conf, names)

    def get_volume_extra_specs(self, volume):
        if volume is None:
            return {}
        type_id = volume.get('volume_type_id', None)
        if type_id is None:
            return {}

        return volume_types.get_volume_type_extra_specs(type_id)

    def require_target_existed(self, targets):
        """Check if the target list includes one or more members."""
        if not targets['list']:
            msg = self.output_log(MSG.NO_CONNECTED_TARGET)
            self.raise_error(msg)

    def raise_error(self, msg):
        """Raise a VolumeDriverException by driver error message."""
        message = _(
            '%(prefix)s error occurred. %(msg)s' % {
                'prefix': self.driver_info['driver_prefix'],
                'msg': msg,
            }
        )
        raise exception.VolumeDriverException(message)

    def raise_busy(self):
        """Raise a VolumeDriverException by driver busy message."""
        message = _(utils.BUSY_MESSAGE)
        raise exception.VolumeDriverException(message)

    def is_controller(self, connector):
        return True if (
            'ip' in connector and connector['ip'] == CONF.my_ip) else False



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Update reservations resource

Revision ID: b8660621f1b9
Revises: 89aa6f9639f9
Create Date: 2021-10-27 17:25:16.790525
"""

from alembic import op
from oslo_log import log as logging
import sqlalchemy as sa


LOG = logging.getLogger(__name__)


# revision identifiers, used by Alembic.
revision = 'b8660621f1b9'
down_revision = '89aa6f9639f9'
branch_labels = None
depends_on = None


def upgrade():
    connection = op.get_bind()

    for table_name in ('quotas', 'quota_classes', 'reservations'):
        table = sa.Table(table_name, sa.MetaData(), autoload_with=connection)
        col = table.c.resource
        # SQLite doesn't support altering tables, so we use a workaround
        if connection.engine.name == 'sqlite':
            with op.batch_alter_table(table_name) as batch_op:
                batch_op.alter_column('resource',
                                      existing_type=col.type,
                                      type_=sa.String(length=300))

        else:
            # MySQL ALTER needs to have existing_type, existing_server_default,
            # and existing_nullable or it will do who-knows-what
            try:
                op.alter_column(table_name, 'resource',
                                existing_type=col.type,
                                existing_nullable=col.nullable,
                                existing_server_default=col.server_default,
                                type_=sa.String(length=300))
            except Exception:
                # On MariaDB, max length varies depending on the version and
                # the InnoDB page size [1], so it is possible to have error
                # 1071 ('Specified key was too long; max key length is 767
                # bytes").  Since this migration is to resolve a corner case,
                # deployments with those DB versions won't be covered.
                # [1]: https://mariadb.com/kb/en/library/innodb-limitations/#page-sizes  # noqa
                if not connection.engine.name == 'mysql':
                    raise
                LOG.warning('Error in migration %s, Cinder still affected by '
                            'bug #1948962', revision)



#    (c)  Copyright 2022 Fungible, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Define all constants required for fungible driver
"""
# API constants
VERSION = '1.0.0'
STATIC_URL = '/FunCC/v1'

# Volume type constants
VOLUME_TYPE_EC = 'VOL_TYPE_BLK_EC'
VOLUME_TYPE_REPLICA = 'VOL_TYPE_BLK_REPLICA'
VOLUME_TYPE_RAW = 'VOL_TYPE_BLK_LOCAL_THIN'
VOLUME_TYPE_RF1 = 'VOL_TYPE_BLK_RF1'

# General constants
FALSE = 'false'
TRUE = 'true'
BOOLEAN = [TRUE, FALSE]
BYTES_PER_GIB = 1073741824
FSC_IOPS_IMG_MIG = "iops_for_image_migration"

# Extra specs constants
FSC_QOS_BAND = 'fungible:qos_band'
FSC_SPACE_ALLOCATION_POLICY = 'fungible:space_allocation_policy'
FSC_COMPRESSION = 'fungible:compression'
FSC_EC_SCHEME = 'fungible:ec_scheme'
FSC_SNAPSHOTS = "fungible:snapshots"
FSC_KMIP_SECRET_KEY = 'fungible:kmip_secret_key'
FSC_VOL_TYPE = 'fungible:vol_type'
FSC_BLK_SIZE = "fungible:block_size"
FSC_FD_IDS = 'fungible:fault_domain_ids'
FSC_FD_OP = 'fungible:fd_op'
BLOCK_SIZE_4K = '4096'
BLOCK_SIZE_8K = '8192'
BLOCK_SIZE_16K = '16384'
BLOCK_SIZE = [BLOCK_SIZE_4K, BLOCK_SIZE_8K, BLOCK_SIZE_16K]
FSC_FD_OPS = ['SUGGESTED_FD_IDS', 'EXCLUDE_FD_IDS', 'ASSIGNED_FD_ID']
SPACE_ALLOCATION_POLICY = ['balanced', 'write_optimized', 'capacity_optimized']
EC_8_2 = '8_2'
EC_4_2 = '4_2'
EC_2_1 = '2_1'
QOS_BAND = {
    'gold': 0,
    'silver': 1,
    'bronze': 2
}



# Copyright (c) 2025 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from unittest import mock

import ddt

from cinder import context
from cinder import exception
from cinder.tests.unit import fake_volume
from cinder.tests.unit.volume.drivers.dell_emc import powerflex


class DictToObject:
    def __init__(self, dictionary):
        for key, value in dictionary.items():
            setattr(self, key, value)

    def get(self, key):
        return self.__dict__.get(key)


@ddt.ddt
class TestSDC(powerflex.TestPowerFlexDriver):
    def setUp(self):
        """Setup a test case environment."""

        super(TestSDC, self).setUp()

        self.client_mock = mock.MagicMock()
        self.driver._get_client = mock.MagicMock(return_value=self.client_mock)

        self.connector = {
            "sdc_guid": "028888FA-502A-4FAC-A888-1FA3B256358C",
            "host": "hostname"}
        self.host_id = "13bf228a00010001"
        self.host = {"name": "hostname"}
        self.ctx = (
            context.RequestContext('fake', 'fake', True, auth_token=True))

        self.attachment1 = DictToObject(fake_volume.fake_db_volume_attachment(
            **{
                'attach_status': 'attached',
                'attached_host': self.host['name']
            }
        ))
        self.attachment2 = DictToObject(fake_volume.fake_db_volume_attachment(
            **{
                'attach_status': 'attached',
                'attached_host': self.host['name']
            }
        ))
        self.volume = fake_volume.fake_volume_obj(
            self.ctx, **{'provider_id': '3bd1f78800000019',
                         'size': 8})

    def test_initialize_connection(self):
        self.driver._initialize_connection = mock.MagicMock()

        self.driver.initialize_connection(self.volume, self.connector)

        self.driver._initialize_connection.assert_called_once_with(
            self.volume, self.connector, self.volume.size)

    def test__initialize_connection(self):
        self.client_mock.query_sdc_id_by_guid.return_value = self.host_id
        self.driver._attach_volume_to_host = mock.MagicMock()
        self.driver._check_volume_mapped = mock.MagicMock()

        result = self.driver._initialize_connection(
            self.volume, self.connector, self.volume.size)

        self.assertEqual(result['driver_volume_type'], "scaleio")
        self.driver._attach_volume_to_host.assert_called_with(
            self.volume, self.host_id)
        self.driver._check_volume_mapped.assert_called_with(
            self.host_id, self.volume.provider_id)

    def test__initialize_connection_no_connector(self):
        self.assertRaises(exception.InvalidHost,
                          self.driver._initialize_connection,
                          self.volume,
                          {},
                          self.volume.size)

    def test__attach_volume_to_host_success(self):
        self.client_mock.query_sdc_by_id.return_value = self.host
        self.client_mock.query_volume.return_value = {
            "mappedSdcInfo": []
        }
        self.client_mock.map_volume.return_value = None

        self.driver._attach_volume_to_host(self.volume, self.host_id)

        self.client_mock.query_sdc_by_id.assert_called_once_with(
            self.host_id)
        self.client_mock.query_volume.assert_called_once_with(
            self.volume.provider_id)
        self.client_mock.map_volume.assert_called_once_with(
            self.volume.provider_id, self.host_id)

    def test__attach_volume_to_host_already_attached(self):
        self.client_mock.query_sdc_by_id.return_value = self.host
        self.client_mock.query_volume.return_value = {
            "mappedSdcInfo": [
                {
                    "sdcId": self.host_id
                }
            ]
        }

        self.driver._attach_volume_to_host(self.volume, self.host_id)

        self.client_mock.query_sdc_by_id.assert_called_once_with(
            self.host_id)
        self.client_mock.map_volume.assert_not_called()

    def test__check_volume_mapped_success(self):
        self.client_mock.query_sdc_volumes.return_value = [
            'vol1', 'vol2', self.volume.id]

        self.driver._check_volume_mapped(self.host_id, self.volume.id)

        self.client_mock.query_sdc_volumes.assert_called_once_with(
            self.host_id)

    def test__check_volume_mapped_fail(self):
        self.client_mock.query_sdc_volumes.return_value = []

        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver._check_volume_mapped,
                          self.host_id, self.volume.id)

    def test__check_volume_mapped_with_retry(self):
        self.client_mock.query_sdc_volumes.side_effect = [
            [],
            [self.volume.id]
        ]
        self.driver._check_volume_mapped('sdc_id', self.volume.id)
        self.assertEqual(self.client_mock.query_sdc_volumes.call_count, 2)

    def test_terminate_connection(self):
        self.driver._terminate_connection = mock.MagicMock()

        self.driver.terminate_connection(self.volume, self.connector)

        self.driver._terminate_connection.assert_called_once_with(
            self.volume, self.connector)

    def test__terminate_connection_success(self):
        self.client_mock.query_sdc_id_by_guid.return_value = self.host_id
        self.driver._detach_volume_from_host = mock.MagicMock(
            return_valure=None)

        self.driver._terminate_connection(self.volume, self.connector)

        self.client_mock.query_sdc_id_by_guid.assert_called_once_with(
            self.connector["sdc_guid"])
        self.driver._detach_volume_from_host.assert_called_once_with(
            self.volume, self.host_id)

    def test__terminate_connection_no_connector(self):
        self.assertRaises(exception.InvalidHost,
                          self.driver._terminate_connection,
                          self.volume,
                          {})

    def test__terminate_connection_multiattached(self):
        self.driver._is_multiattached_to_host = mock.MagicMock(
            return_valure=False)

        self.driver._terminate_connection(self.volume, self.connector)

        self.client_mock.query_sdc_id_by_guid.assert_not_called()

    def test__is_multiattached_to_host_false(self):

        result = self.driver._is_multiattached_to_host(
            [self.attachment1], self.host['name'])

        self.assertFalse(result)

    def test__is_multiattached_to_host_true(self):

        result = self.driver._is_multiattached_to_host(
            [self.attachment1, self.attachment2],
            self.host['name'])

        self.assertTrue(result)

    @ddt.data("13bf228a00010001", None)
    def test__detach_volume_from_host_detached_1(self, host_id):
        self.client_mock.query_volume.return_value = {
            "mappedSdcInfo": []
        }

        self.driver._detach_volume_from_host(self.volume, host_id)

        self.client_mock.unmap_volume.assert_not_called()

    def test__detach_volume_from_host_detached_2(self):
        self.client_mock.query_volume.return_value = {
            "mappedSdcInfo": [
                {
                    "sdcId": "fake_id"
                }
            ]
        }
        self.client_mock.query_sdc_by_id.return_value = self.host

        self.driver._detach_volume_from_host(self.volume, self.host_id)

        self.client_mock.query_sdc_by_id.assert_called_once_with(self.host_id)
        self.client_mock.unmap_volume.assert_not_called()

    def test__detach_volume_from_host_with_hostid(self):
        self.client_mock.query_volume.return_value = {
            "mappedSdcInfo": [
                {
                    "sdcId": self.host_id
                }
            ]
        }
        self.client_mock.query_sdc_by_id.return_value = self.host

        self.driver._detach_volume_from_host(self.volume, self.host_id)

        self.client_mock.query_sdc_by_id.assert_called_once_with(self.host_id)
        self.client_mock.unmap_volume.assert_called_once_with(
            self.volume.provider_id, self.host_id)

    def test__detach_volume_from_host_without_hostid(self):
        self.client_mock.query_volume.return_value = {
            "mappedSdcInfo": [
                {
                    "sdcId": self.host_id
                }
            ]
        }
        self.client_mock.query_sdc_by_id.return_value = self.host

        self.driver._detach_volume_from_host(self.volume)

        self.client_mock.query_sdc_by_id.assert_not_called()
        self.client_mock.unmap_volume.assert_called_once_with(
            self.volume.provider_id)

    def test__check_volume_unmapped_success(self):
        self.client_mock.query_sdc_volumes.return_value = []

        self.driver._check_volume_unmapped(self.host_id, self.volume.id)

        self.client_mock.query_sdc_volumes.assert_called_once_with(
            self.host_id)

    def test__check_volume_unmapped_fail(self):
        self.client_mock.query_sdc_volumes.return_value = [
            'vol1', 'vol2', self.volume.id]

        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver._check_volume_unmapped,
                          self.host_id, self.volume.id)

    def test__check_volume_unmapped_with_retry(self):
        self.client_mock.query_sdc_volumes.side_effect = [
            [self.volume.id],
            []
        ]
        self.driver._check_volume_unmapped('sdc_id', self.volume.id)
        self.assertEqual(self.client_mock.query_sdc_volumes.call_count, 2)



# Copyright (c) 2020 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from cinder import exception
from cinder.objects import fields
from cinder.objects import volume_attachment
from cinder.tests.unit import fake_constants
from cinder.tests.unit import fake_volume
from cinder.tests.unit.volume.drivers.dell_emc import powerstore
from cinder.volume.drivers.dell_emc.powerstore import utils

FAKE_HOST = {
    "name": "fake_host",
    "id": "fake_id"
}


class TestVolumeAttachDetach(powerstore.TestPowerStoreDriver):

    QOS_SPECS = {
        'qos_specs': {
            'name': 'powerstore_qos',
            'id': 'd8c88f5a-4c6f-4f89-97c5-da1ef059006e',
            'created_at': 'fake_date',
            'consumer': 'back-end',
            'specs': {
                'max_bw': '104857600',
                'max_iops': '500',
                'bandwidth_limit_type': 'Absolute',
                'burst_percentage': '50'
            }
        }
    }

    @mock.patch("cinder.volume.drivers.dell_emc.powerstore.client."
                "PowerStoreClient.get_chap_config")
    def setUp(self, mock_chap):
        super(TestVolumeAttachDetach, self).setUp()
        mock_chap.return_value = {"mode": "Single"}
        self.iscsi_driver.check_for_setup_error()
        self.fc_driver.check_for_setup_error()
        with mock.patch.object(self.nvme_driver.adapter.client,
                               "get_array_version",
                               return_value=(
                                   "3.0.0.0"
                               )):
            self.nvme_driver.check_for_setup_error()
        self.volume = fake_volume.fake_volume_obj(
            self.context,
            host="host@backend",
            provider_id="fake_id",
            size=8,
            volume_type_id=fake_constants.VOLUME_TYPE_ID
        )
        self.volume.volume_attachment = (
            volume_attachment.VolumeAttachmentList()
        )
        self.volume.volume_attachment.objects = [
            volume_attachment.VolumeAttachment(
                attach_status=fields.VolumeAttachStatus.ATTACHED,
                attached_host=self.volume.host
            ),
            volume_attachment.VolumeAttachment(
                attach_status=fields.VolumeAttachStatus.ATTACHED,
                attached_host=self.volume.host
            )
        ]
        fake_iscsi_targets_response = [
            {
                "address": "1.2.3.4",
                "ip_port": {
                    "target_iqn":
                        "iqn.2020-07.com.dell:dellemc-powerstore-test-iqn-1"
                },
            },
            {
                "address": "5.6.7.8",
                "ip_port": {
                    "target_iqn":
                        "iqn.2020-07.com.dell:dellemc-powerstore-test-iqn-1"
                },
            },
        ]
        fake_fc_wwns_response = [
            {
                "wwn": "58:cc:f0:98:49:21:07:02"
            },
            {
                "wwn": "58:cc:f0:98:49:23:07:02"
            },
        ]
        fake_nvme_portals_response = [
            {
                "address": "11.22.33.44"
            },
            {
                "address": "55.66.77.88"
            }
        ]
        fake_nvme_nqn_response = [
            {
                "nvm_subsystem_nqn":
                    "nqn.2020-07.com.dell:powerstore:00:test-nqn"
            }
        ]
        self.fake_connector = {
            "host": self.volume.host,
            "wwpns": ["58:cc:f0:98:49:21:07:02", "58:cc:f0:98:49:23:07:02"],
            "initiator": "fake_initiator",
        }
        self.iscsi_targets_mock = self.mock_object(
            self.iscsi_driver.adapter.client,
            "get_ip_pool_address",
            return_value=fake_iscsi_targets_response
        )
        self.fc_wwns_mock = self.mock_object(
            self.fc_driver.adapter.client,
            "get_fc_port",
            return_value=fake_fc_wwns_response
        )
        self.nvme_portal_mock = self.mock_object(
            self.nvme_driver.adapter.client,
            "get_ip_pool_address",
            return_value=fake_nvme_portals_response
        )
        self.nvme_nqn_mock = self.mock_object(
            self.nvme_driver.adapter.client,
            "get_subsystem_nqn",
            return_value=fake_nvme_nqn_response
        )

    def test_initialize_connection_chap_enabled(self):
        self.iscsi_driver.adapter.use_chap_auth = True
        with mock.patch.object(self.iscsi_driver.adapter,
                               "_create_host_and_attach",
                               return_value=(
                                   utils.get_chap_credentials(),
                                   1
                               )):
            connection_properties = self.iscsi_driver.initialize_connection(
                self.volume,
                self.fake_connector
            )
            self.assertIn("auth_username", connection_properties["data"])
            self.assertIn("auth_password", connection_properties["data"])

    def test_initialize_connection_chap_disabled(self):
        self.iscsi_driver.adapter.use_chap_auth = False
        with mock.patch.object(self.iscsi_driver.adapter,
                               "_create_host_and_attach",
                               return_value=(
                                   utils.get_chap_credentials(),
                                   1
                               )):
            connection_properties = self.iscsi_driver.initialize_connection(
                self.volume,
                self.fake_connector
            )
            self.assertNotIn("auth_username", connection_properties["data"])
            self.assertNotIn("auth_password", connection_properties["data"])

    def test_get_fc_targets(self):
        wwns = self.fc_driver.adapter._get_fc_targets()
        self.assertEqual(2, len(wwns))

    def test_get_fc_targets_filtered(self):
        self.fc_driver.adapter.allowed_ports = ["58:cc:f0:98:49:23:07:02"]
        wwns = self.fc_driver.adapter._get_fc_targets()
        self.assertEqual(1, len(wwns))
        self.assertNotIn(
            utils.fc_wwn_to_string("58:cc:f0:98:49:21:07:02"), wwns
        )

    def test_get_fc_targets_filtered_no_matched_ports(self):
        self.fc_driver.adapter.allowed_ports = ["fc_wwn_1", "fc_wwn_2"]
        error = self.assertRaises(exception.VolumeBackendAPIException,
                                  self.fc_driver.adapter._get_fc_targets)
        self.assertIn("There are no accessible Fibre Channel targets on the "
                      "system.", error.msg)

    def test_get_iscsi_targets(self):
        iqns, portals = self.iscsi_driver.adapter._get_iscsi_targets()
        self.assertTrue(len(iqns) == len(portals))
        self.assertEqual(2, len(portals))

    def test_get_iscsi_targets_filtered(self):
        self.iscsi_driver.adapter.allowed_ports = ["1.2.3.4"]
        iqns, portals = self.iscsi_driver.adapter._get_iscsi_targets()
        self.assertTrue(len(iqns) == len(portals))
        self.assertEqual(1, len(portals))
        self.assertNotIn(
            "iqn.2020-07.com.dell:dellemc-powerstore-test-iqn-2", iqns
        )

    def test_get_iscsi_targets_filtered_no_matched_ports(self):
        self.iscsi_driver.adapter.allowed_ports = ["1.1.1.1", "2.2.2.2"]
        error = self.assertRaises(exception.VolumeBackendAPIException,
                                  self.iscsi_driver.adapter._get_iscsi_targets)
        self.assertIn("There are no accessible iSCSI targets on the system.",
                      error.msg)

    def test_get_nvme_targets(self):
        portals, nqn = self.nvme_driver.adapter._get_nvme_targets()
        self.assertEqual(2, len(portals))

    def test_get_connection_properties(self):
        volume_identifier = '123'
        portals, nqn = self.nvme_driver.adapter._get_nvme_targets()
        result = {
            'driver_volume_type': 'nvmeof',
            'data': {
                'portals': [('11.22.33.44', 4420, 'tcp'),
                            ('55.66.77.88', 4420, 'tcp')],
                'target_nqn': [{
                    'nvm_subsystem_nqn':
                        'nqn.2020-07.com.dell:powerstore:00:test-nqn'
                }],
                'volume_nguid': '123',
                'discard': True
            }
        }
        self.assertEqual(result,
                         self.nvme_driver.adapter.
                         _get_connection_properties(volume_identifier))

    def test_get_connection_properties_no_volume_identifier(self):
        portals, nqn = self.nvme_driver.adapter._get_nvme_targets()
        result = {
            'driver_volume_type': 'nvmeof',
            'data': {
                'portals': [('11.22.33.44', 4420, 'tcp'),
                            ('55.66.77.88', 4420, 'tcp')],
                'target_nqn': [{
                    'nvm_subsystem_nqn':
                        'nqn.2020-07.com.dell:powerstore:00:test-nqn'
                }],
                'volume_nguid': None,
                'discard': True
            }
        }
        self.assertEqual(result, self.nvme_driver.adapter.
                         _get_connection_properties(None))

    def test_get_connection_properties_no_nqn(self):
        volume_identifier = '123'
        with mock.patch.object(self.nvme_driver.adapter,
                               "_get_nvme_targets",
                               return_value=(['11.22.33.44', '55.66.77.88'],
                                             [])):
            result = {
                'driver_volume_type': 'nvmeof',
                'data': {
                    'portals': [('11.22.33.44', 4420, 'tcp'),
                                ('55.66.77.88', 4420, 'tcp')],
                    'target_nqn': [],
                    'volume_nguid': '123',
                    'discard': True
                }
            }
            self.assertEqual(result, self.nvme_driver.adapter.
                             _get_connection_properties(volume_identifier))

    def test_get_connection_properties_no_portals(self):
        volume_identifier = '123'
        with mock.patch.object(self.nvme_driver.adapter,
                               "_get_nvme_targets",
                               return_value=(
                                   [],
                                   [{
                                       'nvm_subsystem_nqn':
                                           'nqn.2020-07.com.dell:powerstore:0'
                                           '0:test-nqn'
                                   }]
                               )):
            result = {
                'driver_volume_type': 'nvmeof',
                'data': {
                    'portals': [],
                    'target_nqn': [{
                        'nvm_subsystem_nqn':
                        'nqn.2020-07.com.dell:powerstore:00:test-nqn'
                    }],
                    'volume_nguid': '123',
                    'discard': True
                }
            }
            self.assertEqual(result, self.nvme_driver.adapter.
                             _get_connection_properties(volume_identifier))

    @mock.patch("cinder.volume.drivers.dell_emc.powerstore.adapter."
                "CommonAdapter._detach_volume_from_hosts")
    @mock.patch("cinder.volume.drivers.dell_emc.powerstore.adapter."
                "CommonAdapter._filter_hosts_by_initiators")
    def test_detach_multiattached_volume(self, mock_filter_hosts, mock_detach):
        self.iscsi_driver.terminate_connection(self.volume,
                                               self.fake_connector)
        mock_filter_hosts.assert_not_called()
        mock_detach.assert_not_called()
        self.volume.volume_attachment.objects.pop()
        self.iscsi_driver.terminate_connection(self.volume,
                                               self.fake_connector)
        mock_filter_hosts.assert_called_once()
        mock_detach.assert_called_once()

    @mock.patch('cinder.volume.volume_types.'
                'get_volume_type_qos_specs',
                return_value=QOS_SPECS)
    @mock.patch("cinder.volume.drivers.dell_emc.powerstore.client."
                "PowerStoreClient.attach_volume_to_host")
    @mock.patch("cinder.volume.drivers.dell_emc.powerstore.client."
                "PowerStoreClient.get_volume_lun")
    @mock.patch("cinder.volume.drivers.dell_emc.powerstore.client."
                "PowerStoreClient.get_array_version",
                return_value='4.0')
    @mock.patch("cinder.volume.drivers.dell_emc.powerstore.client."
                "PowerStoreClient.get_qos_policy_id_by_name",
                return_value=None)
    @mock.patch("cinder.volume.drivers.dell_emc.powerstore.client."
                "PowerStoreClient.create_qos_io_rule")
    @mock.patch("cinder.volume.drivers.dell_emc.powerstore.client."
                "PowerStoreClient.create_qos_policy")
    @mock.patch("cinder.volume.drivers.dell_emc.powerstore.client."
                "PowerStoreClient.update_volume_with_qos_policy")
    def test_volume_qos_policy_create(self,
                                      mock_volume_types,
                                      mock_attach_volume,
                                      mock_get_volume_lun,
                                      mock_get_array_version,
                                      mock_get_qos_policy,
                                      mock_qos_io_rule,
                                      mock_qos_policy,
                                      mock_volume_qos_update):

        self.iscsi_driver.adapter.use_chap_auth = False
        self.mock_object(self.iscsi_driver.adapter,
                         "_create_host_if_not_exist",
                         return_value=(
                             FAKE_HOST,
                             utils.get_chap_credentials(),
                         ))
        self.iscsi_driver.initialize_connection(
            self.volume,
            self.fake_connector
        )
        mock_get_volume_lun.return_value = "fake_volume_identifier"
        mock_qos_io_rule.return_value = "9beb10ff-a00c-4d88-a7d9-692be2b3073f"
        mock_qos_policy.return_value = "d69f7131-4617-4bae-89f8-a540a6bda94b"
        mock_volume_types.assert_called_once()
        mock_get_array_version.assert_called_once()
        mock_get_qos_policy.assert_called_once()
        mock_attach_volume.assert_called_once()
        mock_qos_policy.assert_called_once()
        mock_volume_qos_update.assert_called_once()

    @mock.patch('cinder.volume.volume_types.'
                'get_volume_type_qos_specs',
                return_value=QOS_SPECS)
    @mock.patch("cinder.volume.drivers.dell_emc.powerstore.client."
                "PowerStoreClient.attach_volume_to_host")
    @mock.patch("cinder.volume.drivers.dell_emc.powerstore.client."
                "PowerStoreClient.get_volume_lun")
    @mock.patch("cinder.volume.drivers.dell_emc.powerstore.client."
                "PowerStoreClient.get_array_version",
                return_value='4.0')
    @mock.patch("cinder.volume.drivers.dell_emc.powerstore.client."
                "PowerStoreClient.get_qos_policy_id_by_name")
    @mock.patch("cinder.volume.drivers.dell_emc.powerstore.client."
                "PowerStoreClient.update_qos_io_rule")
    @mock.patch("cinder.volume.drivers.dell_emc.powerstore.client."
                "PowerStoreClient.update_volume_with_qos_policy")
    def test_volume_qos_io_rule_update(self,
                                       mock_volume_types,
                                       mock_attach_volume,
                                       mock_get_volume_lun,
                                       mock_get_array_version,
                                       mock_get_qos_policy,
                                       mock_update_qos_io_rule,
                                       mock_volume_qos_update):
        self.iscsi_driver.adapter.use_chap_auth = False
        self.mock_object(self.iscsi_driver.adapter,
                         "_create_host_if_not_exist",
                         return_value=(
                             FAKE_HOST,
                             utils.get_chap_credentials(),
                         ))
        self.iscsi_driver.initialize_connection(
            self.volume,
            self.fake_connector
        )
        mock_get_volume_lun.return_value = "fake_volume_identifier"
        mock_get_qos_policy.return_value = ("d69f7131-"
                                            "4617-4bae-89f8-a540a6bda94b")
        mock_volume_types.assert_called_once()
        mock_attach_volume.assert_called_once()
        mock_get_array_version.assert_called_once()
        mock_update_qos_io_rule.assert_called_once()
        mock_volume_qos_update.assert_called_once()

    @mock.patch('cinder.volume.volume_types.'
                'get_volume_type_qos_specs',
                return_value=QOS_SPECS)
    @mock.patch("cinder.volume.drivers.dell_emc.powerstore.utils."
                "is_multiattached_to_host", return_value=False)
    @mock.patch("cinder.volume.drivers.dell_emc.powerstore.client."
                "PowerStoreClient.detach_volume_from_host")
    @mock.patch("cinder.volume.drivers.dell_emc.powerstore.client."
                "PowerStoreClient.get_array_version",
                return_value='4.0')
    @mock.patch("cinder.volume.drivers.dell_emc.powerstore.client."
                "PowerStoreClient.update_volume_with_qos_policy")
    def test_volume_qos_policy_update(self,
                                      mock_volume_types,
                                      mock_multi_attached_host,
                                      mock_detach_volume,
                                      mock_get_array_version,
                                      mock_volume_qos_update):
        self.mock_object(self.iscsi_driver.adapter,
                         "_filter_hosts_by_initiators",
                         return_value=FAKE_HOST)
        self.iscsi_driver.terminate_connection(self.volume,
                                               self.fake_connector)
        mock_volume_types.assert_called_once()
        mock_multi_attached_host.assert_called_once()
        mock_detach_volume.assert_called_once()
        mock_get_array_version.assert_called_once()
        mock_volume_qos_update.assert_called_once()



# Copyright 2016 Dell Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""
Core backend volume driver interface.

All backend drivers should support this interface as a bare minimum, but some
methods (marked as optional in their description) can rely on the default
implementation.
"""

from cinder.interface import base


class VolumeDriverCore(base.CinderInterface):
    """Core backend driver required interface."""

    def do_setup(self, context):
        """Any initialization the volume driver needs to do while starting.

        Called once by the manager after the driver is loaded.
        Can be used to set up clients, check licenses, set up protocol
        specific helpers, etc.

        If you choose to raise an exception here, the setup is considered
        failed already and the check_for_setup_error() will not be called.

        :param context: The admin context of type context.RequestContext.
        :raises InvalidConfigurationValue: raise this if you detect a problem
                                           during a configuration check
        :raises VolumeDriverException: raise this or one of its more specific
                                       subclasses if you detect setup problems
                                       other than invalid configuration
        """

    def check_for_setup_error(self):
        """Validate there are no issues with the driver configuration.

        Called after do_setup(). Driver initialization can occur there or in
        this call, but must be complete by the time this returns.

        If this method raises an exception, the driver will be left in an
        "uninitialized" state by the volume manager, which means that it will
        not be sent requests for volume operations.

        This method typically checks things like whether the configured
        credentials can be used to log in the storage backend, and whether any
        external dependencies are present and working.

        :raises VolumeBackendAPIException: in case of setup error.
        :raises InvalidConfigurationValue: raise this if you detect a problem
                                           during a configuration check
        """

    def get_volume_stats(self, refresh=False):
        """Collects volume backend stats.

        The get_volume_stats method is used by the volume manager to collect
        information from the driver instance related to information about the
        driver, available and used space, and driver/backend capabilities.

        stats are stored in 'self._stats' field, which could be updated in
        '_update_volume_stats' method.

        It returns a dict with the following required fields:

        * volume_backend_name
            This is an identifier for the backend taken from cinder.conf.
            Useful when using multi-backend.
        * vendor_name
            Vendor/author of the driver who serves as the contact for the
            driver's development and support.
        * driver_version
            The driver version is logged at cinder-volume startup and is useful
            for tying volume service logs to a specific release of the code.
            There are currently no rules for how or when this is updated, but
            it tends to follow typical major.minor.revision ideas.
        * storage_protocol
            The protocol used to connect to the storage, this should be a short
            string such as: "iSCSI", "FC", "NFS", "ceph", etc.
            Available protocols are present in cinder.common.constants and they
            must be used instead of string literals.
            Variant values only exist for older drivers that were already
            reporting those values.  New drivers must use non variant versions.
            In some cases this may be the same value as the driver_volume_type
            returned by the initialize_connection method, but they are not the
            same thing, since this one is meant to be used by the scheduler,
            while the latter is the os-brick connector identifier used in the
            factory method.

        * total_capacity_gb
            The total capacity in gigabytes (GiB) of the storage backend being
            used to store Cinder volumes. Use keyword 'unknown' if the backend
            cannot report the value or 'infinite' if there is no upper limit.
            But, it is recommended to report real values as the Cinder
            scheduler assigns lowest weight to any storage backend reporting
            'unknown' or 'infinite'.

        * free_capacity_gb
            The free capacity in gigabytes (GiB). Use keyword 'unknown' if the
            backend cannot report the value or 'infinite' if there is no upper
            limit. But, it is recommended to report real values as the Cinder
            scheduler assigns lowest weight to any storage backend reporting
            'unknown' or 'infinite'.

        And the following optional fields:

        * reserved_percentage (integer)
            Percentage of backend capacity which is not used by the scheduler.
        * location_info (string)
            Driver-specific information used by the driver and storage backend
            to correlate Cinder volumes and backend LUNs/files.
        * QoS_support (Boolean)
            Whether the backend supports quality of service.
        * provisioned_capacity_gb
            The total provisioned capacity on the storage backend, in gigabytes
            (GiB), including space consumed by any user other than Cinder
            itself.
        * max_over_subscription_ratio
            The maximum amount a backend can be over subscribed.
        * thin_provisioning_support (Boolean)
            Whether the backend is capable of allocating thinly provisioned
            volumes.
        * thick_provisioning_support (Boolean)
            Whether the backend is capable of allocating thick provisioned
            volumes. (Typically True.)
        * total_volumes (integer)
            Total number of volumes on the storage backend. This can be used in
            custom driver filter functions.
        * filter_function (string)
            A custom function used by the scheduler to determine whether a
            volume should be allocated to this backend or not. Example:

              capabilities.total_volumes < 10

        * goodness_function (string)
            Similar to filter_function, but used to weigh multiple volume
            backends. Example:

              capabilities.capacity_utilization < 0.6 ? 100 : 25

        * multiattach (Boolean)
            Whether the backend supports multiattach or not. Defaults to False.
        * sparse_copy_volume (Boolean)
            Whether copies performed by the volume manager for operations such
            as migration should attempt to preserve sparseness.
        * online_extend_support (Boolean)
            Whether the backend supports in-use volume extend or not. Defaults
            to True.
        * clone_across_pools (Boolean)
            Whether the backend supports cloning a volume across different
            pools. Defaults to False.

        The returned dict may also contain a list, "pools", which has a similar
        dict for each pool being used with the backend.

        :param refresh: Whether to discard any cached values and force a full
                        refresh of stats.
        :returns: dict of appropriate values (see above).
        """

    def create_volume(self, volume):
        """Create a new volume on the backend.

        This method is responsible only for storage allocation on the backend.
        It should not export a LUN or actually make this storage available for
        use, this is done in a later call.

        TODO(smcginnis): Add example data structure of volume object.

        :param volume: Volume object containing specifics to create.
        :returns: (Optional) dict of database updates for the new volume.
        :raises VolumeBackendAPIException: if creation failed.
        """

    def delete_volume(self, volume):
        """Delete a volume from the backend.

        If the driver can talk to the backend and detects that the volume is no
        longer present, this call should succeed and allow Cinder to complete
        the process of deleting the volume.

        It is imperative that this operation ensures that the data from the
        deleted volume cannot leak into new volumes when they are created, as
        new volumes are likely to belong to a different tenant/project.

        :param volume: The volume to delete.
        :raises VolumeIsBusy: if the volume is still attached or has snapshots.
                 VolumeBackendAPIException on error.
        """

    def initialize_connection(self, volume, connector, initiator_data=None):
        """Allow connection to connector and return connection info.

        :param volume: The volume to be attached.
        :param connector: Dictionary containing information about what is being
                          connected to.
        :param initiator_data: (Optional) A dictionary of driver_initiator_data
                               objects with key-value pairs that have been
                               saved for this initiator by a driver in previous
                               initialize_connection calls.
        :returns: A dictionary of connection information. This can optionally
                  include a "initiator_updates" field.

        The "initiator_updates" field must be a dictionary containing a
        "set_values" and/or "remove_values" field. The "set_values" field must
        be a dictionary of key-value pairs to be set/updated in the db. The
        "remove_values" field must be a list of keys, previously set with
        "set_values", that will be deleted from the db.

        May be called multiple times to get connection information after a
        volume has already been attached.
        """

    def terminate_connection(self, volume, connector):
        """Remove access to a volume.

        Note: If ``connector`` is ``None``, then all connections to the volume
        should be terminated.

        :param volume: The volume to remove.
        :param connector: The Dictionary containing information about the
                          connection. This is optional when doing a
                          force-detach and can be None.
        """

    def clone_image(self, context, volume,
                    image_location, image_meta, image_service):
        """Create a volume efficiently from an existing image.

        Drivers that, always or under some circumstances, can efficiently
        create a volume from a Glance image can implement this method to be
        given a chance to try to do the volume creation as efficiently as
        possible.

        If the driver cannot do it efficiently on a specific call it can
        return ``(None, False)`` to let Cinder try other mechanisms.

        **This method is optional** and most drivers won't need to implement
        it and can leverage the default driver implementation that returns
        ``(None, False)`` to indicate that this optimization is not possible on
        this driver.

        Examples where drivers can do this optimization:

        - When images are stored on the same storage system and the driver can
          locate them and efficiently create a volume.  For example the RBD
          driver can efficiently create a volume if the image is stored on the
          same Ceph cluster and the image format is ``raw``.  Another example
          is the GPFS driver.

        - When volumes are locally accessible and accessing them that way is
          more efficient than going through the remote connection mechanism.
          For example in the GPFS driver if the cloning feature doesn't work it
          will copy the file without using os-brick to connect to the volume.

        :param context: Security/policy info for the request.
        :param volume: The volume to create, as an OVO instance. Drivers should
                       use attributes to access its values instead of using the
                       dictionary compatibility interface it provides.
        :param image_location: Tuple with (``direct_url``, ``locations``) from
                               the `image metadata fields.
                               <https://docs.openstack.org/api-ref/image/v2/index.html?expanded=show-image-detail#show-image-detail>`_
                               ``direct_url``, when present, is a string whose
                               format depends on the image service's external
                               storage in use.
                               Any, or both, tuple positions can be None,
                               depending on the image service configuration.
                               ``locations``, when present, is a list of
                               dictionaries where the value of the ``url`` key
                               contains the direct urls (including the one from
                               ``direct_url``).
        :param image_meta: Dictionary containing `information about the image
                           <https://docs.openstack.org/api-ref/image/v2/index.html?expanded=show-image-detail#show-image-detail>`_,
                           including basic attributes and custom properties.
                           Some transformations have been applied, such as
                           converting timestamps (from ``created_at``,
                           ``updated_at``, and ``deleted_at``) to datetimes,
                           and deserializing JSON values from
                           ``block_device_mapping`` and ``mappings`` keys if
                           present.
                           Base properties, as per the image's schema, will be
                           stored on the base dictionary and the rest will be
                           stored under the ``properties`` key.
                           An important field to check in this method is the
                           ``disk_format`` (e.g.  raw, qcow2).
        :param image_service: The image service to use (``GlanceImageService``
                              instance).  Can fetch image data directly using
                              it.
        :returns: Tuple of (model_update, boolean) where the boolean specifies
                  whether the clone occurred.
        """

    def copy_image_to_volume(self, context, volume, image_service, image_id,
                             disable_sparse=False):
        """Fetch the image from image_service and write it to the volume.

        :param context: Security/policy info for the request.
        :param volume: The volume to create.
        :param image_service: The image service to use.
        :param image_id: The image identifier.
        :param disable_sparse: Enable or disable sparse copy. Default=False.
        :returns: Model updates.
        """

    def copy_volume_to_image(self, context, volume, image_service, image_meta):
        """Copy the volume to the specified image.

        :param context: Security/policy info for the request.
        :param volume: The volume to copy.
        :param image_service: The image service to use.
        :param image_meta: Information about the image.
        :returns: Model updates.
        """

    def before_volume_copy(self, context, src_vol, dest_vol, remote=None):
        """Driver-specific actions executed before copying a volume.

        This method will be called before _copy_volume_data during volume
        migration.

        :param context: Context
        :param src_volume: Source volume in the copy operation.
        :param dest_volume: Destination volume in the copy operation.
        :param remote: Whether the copy operation is local.
        :returns: There is no return value for this method.
        """

    def after_volume_copy(self, context, src_vol, dest_vol, remote=None):
        """Driver-specific actions executed after copying a volume.

        This method will be called after _copy_volume_data during volume
        migration.

        :param context: Context
        :param src_volume: Source volume in the copy operation.
        :param dest_volume: Destination volume in the copy operation.
        :param remote: Whether the copy operation is local.
        :returns: There is no return value for this method.
        """

    def extend_volume(self, volume, new_size):
        """Extend the size of a volume.

        :param volume: The volume to extend.
        :param new_size: The new desired size of the volume.

        Note that if the volume backend doesn't support extending an in-use
        volume, the driver should report online_extend_support=False.
        """

    def migrate_volume(self, context, volume, host):
        """Migrate the volume to the specified host.

        :param context: Context
        :param volume: A dictionary describing the volume to migrate
        :param host: A dictionary describing the host to migrate to, where
                     host['host'] is its name, and host['capabilities'] is a
                     dictionary of its reported capabilities.
        :returns: Tuple of (model_update, boolean) where the boolean specifies
                  whether the migration occurred.
        """

    def update_migrated_volume(self, context, volume, new_volume,
                               original_volume_status):
        """Return model update for migrated volume.

        Each driver implementing this method needs to be responsible for the
        values of _name_id and provider_location. If None is returned or either
        key is not set, it means the volume table does not need to change the
        value(s) for the key(s).
        The return format is {"_name_id": value, "provider_location": value}.

        :param context: Context
        :param volume: The original volume that was migrated to this backend
        :param new_volume: The migration volume object that was created on
                           this backend as part of the migration process
        :param original_volume_status: The status of the original volume
        :returns: model_update to update DB with any needed changes
        """

    def retype(self, context, volume, new_type, diff, host):
        """Change the type of a volume.

        This operation occurs on the same backend and the return value
        indicates whether it was successful.  If migration is required
        to satisfy a retype, that will be handled by the volume manager.

        :param context: Context
        :param volume: The volume to retype
        :param new_type: The target type for the volume
        :param diff: The differences between the two types
        :param host: The host that contains this volume
        :returns: Tuple of (boolean, model_update) where the boolean specifies
                  whether the retype occurred.
        """

    def create_snapshot(self, snapshot):
        """Creates a snapshot.

        :param snapshot: Information for the snapshot to be created.
        """

    def delete_snapshot(self, snapshot):
        """Deletes a snapshot.

        :param snapshot: The snapshot to delete.
        """

    def create_volume_from_snapshot(self, volume, snapshot):
        """Creates a volume from a snapshot.

        If volume_type extra specs includes 'replication: <is> True'
        the driver needs to create a volume replica (secondary),
        and setup replication between the newly created volume and
        the secondary volume.

        An optional larger size for the new volume can be specified. Drivers
        should check this value and create or expand the new volume to match.

        :param volume: The volume to be created.
        :param snapshot: The snapshot from which to create the volume.
        :returns: A dict of database updates for the new volume.
        """

    def set_initialized(self):
        """Mark driver as initialized.

        Do not implement this in a driver. Rely on the default implementation.
        """

    def initialized(self):
        """Getter for driver's initialized status.

        Do not implement this in a driver. Rely on the default implementation.
        """

    def supported(self):
        """Getter for driver's supported status.

        Do not implement this in a driver. Rely on the default implementation.
        """

    def set_throttle(self):
        """Hook for initialization of cinder.volume.throttle.

        This has not been necessary to re-implement or override in any
        drivers thus far. The generic implementation does nothing unless
        explicitly enabled.
        """

    def init_capabilities(self):
        """Fetch and merge capabilities of the driver.

        Do not override this, implement _init_vendor_properties instead.
        """

    def _init_vendor_properties(self):
        """Create a dictionary of vendor unique properties.

        Compose a dictionary by calling ``self._set_property``.

        Select a prefix from the vendor, product, or device name.
        Prefix must match the part of property name before colon (:).

        :returns tuple (properties: dict, prefix: str)
        """

    def update_provider_info(self, volumes, snapshots):
        """Get provider info updates from driver.

        This retrieves a list of volumes and a list of snapshots that
        changed their providers thanks to the initialization of the host,
        so that Cinder can update this information in the volume database.

        This is only implemented by drivers where such migration is possible.

        :param volumes: List of Cinder volumes to check for updates
        :param snapshots: List of Cinder snapshots to check for updates
        :returns: tuple (volume_updates, snapshot_updates)

        where volume updates {'id': uuid, provider_id: <provider-id>}
        and snapshot updates {'id': uuid, provider_id: <provider-id>}
        """



# Copyright (c) 2013 - 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import urllib.parse

from cinder import context
from cinder import exception
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit.fake_snapshot import fake_snapshot_obj
from cinder.tests.unit.fake_volume import fake_volume_obj
from cinder.tests.unit.volume.drivers.dell_emc import powerflex
from cinder.tests.unit.volume.drivers.dell_emc.powerflex import mocks
from cinder.volume import configuration
from cinder.volume.drivers.dell_emc.powerflex import utils as flex_utils


class TestDeleteSnapShot(powerflex.TestPowerFlexDriver):
    """Test cases for ``PowerFlexDriver.delete_snapshot()``"""
    def setUp(self):
        """Setup a test case environment.

        Creates fake volume and snapshot objects and sets up the required
        API responses.
        """
        super(TestDeleteSnapShot, self).setUp()
        ctx = context.RequestContext('fake', 'fake', auth_token=True)

        self.fake_volume = fake_volume_obj(
            ctx, **{'provider_id': fake.PROVIDER_ID})

        self.snapshot = fake_snapshot_obj(
            ctx, **{'volume': self.fake_volume,
                    'provider_id': fake.SNAPSHOT_ID})

        self.snapshot_name_2x_enc = urllib.parse.quote(
            urllib.parse.quote(
                flex_utils.id_to_base64(self.snapshot.id)
            )
        )

        self.HTTPS_MOCK_RESPONSES = {
            self.RESPONSE_MODE.Valid: {
                'types/Volume/instances/getByName::' +
                self.snapshot_name_2x_enc: self.snapshot.id,
                'instances/Volume::' + self.snapshot.provider_id: {},
                'instances/Volume::{}/action/removeMappedSdc'.format(
                    self.snapshot.provider_id
                ): self.snapshot.id,
                'instances/Volume::{}/action/removeVolume'.format(
                    self.snapshot.provider_id
                ): self.snapshot.id,
            },
            self.RESPONSE_MODE.BadStatus: {
                'instances/Volume::' + self.snapshot.provider_id:
                    self.BAD_STATUS_RESPONSE,
                'types/Volume/instances/getByName::' +
                self.snapshot_name_2x_enc: self.BAD_STATUS_RESPONSE,
                'instances/Volume::{}/action/removeVolume'.format(
                    self.snapshot.provider_id
                ): self.BAD_STATUS_RESPONSE,
            },
            self.RESPONSE_MODE.Invalid: {
                'types/Volume/instances/getByName::' +
                self.snapshot_name_2x_enc: mocks.MockHTTPSResponse(
                    {
                        'errorCode': self.OLD_VOLUME_NOT_FOUND_ERROR,
                        'message': 'Test Delete Invalid Snapshot',
                    }, 400
                ),
                'instances/Volume::{}/action/removeVolume'.format(
                    self.snapshot.provider_id): mocks.MockHTTPSResponse(
                    {
                        'errorCode': self.OLD_VOLUME_NOT_FOUND_ERROR,
                        'message': 'Test Delete Invalid Snapshot',
                    }, 400,
                )
            },
        }

    def test_bad_login(self):
        self.set_https_response_mode(self.RESPONSE_MODE.BadStatus)
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.delete_snapshot, self.snapshot)

    def test_delete_invalid_snapshot(self):
        self.set_https_response_mode(self.RESPONSE_MODE.Valid)
        self.driver.delete_snapshot(self.snapshot)

    def test_delete_snapshot(self):
        """Setting the unmap volume before delete flag for tests """
        self.override_config('powerflex_unmap_volume_before_deletion', True,
                             configuration.SHARED_CONF_GROUP)
        self.set_https_response_mode(self.RESPONSE_MODE.Valid)
        self.driver.delete_snapshot(self.snapshot)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from cinder.compute import nova
from cinder import context
from cinder import db
from cinder import exception
from cinder import objects
from cinder.tests.unit.api.v2 import fakes as v2_fakes
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import fake_volume
from cinder.tests.unit import test
from cinder.tests.unit import utils as tests_utils
from cinder.volume import api as volume_api
from cinder.volume import configuration as conf


class AttachmentManagerTestCase(test.TestCase):
    """Attachment related test for volume/api.py."""

    def setUp(self):
        """Setup test class."""
        super(AttachmentManagerTestCase, self).setUp()
        self.configuration = mock.Mock(conf.Configuration)
        self.context = context.get_admin_context()
        self.context.user_id = fake.USER_ID
        self.project_id = fake.PROJECT3_ID
        self.context.project_id = self.project_id
        self.volume_api = volume_api.API()
        self.user_context = context.RequestContext(
            user_id=fake.USER_ID,
            project_id=fake.PROJECT3_ID)

    def test_attachment_create_no_connector(self):
        """Test attachment_create no connector."""
        volume_params = {'status': 'available'}

        vref = tests_utils.create_volume(self.context, **volume_params)
        aref = self.volume_api.attachment_create(self.context,
                                                 vref,
                                                 fake.UUID2)
        self.assertEqual(fake.UUID2, aref.instance_uuid)
        self.assertIsNone(aref.attach_time)
        self.assertEqual('reserved', aref.attach_status)
        self.assertEqual('null', aref.attach_mode)
        self.assertEqual(vref.id, aref.volume_id)
        self.assertEqual({}, aref.connection_info)

    @mock.patch('cinder.volume.rpcapi.VolumeAPI.attachment_update')
    def test_attachment_create_with_connector(self,
                                              mock_rpc_attachment_update):
        """Test attachment_create with connector."""
        volume_params = {'status': 'available'}
        connection_info = {'fake_key': 'fake_value',
                           'fake_key2': ['fake_value1', 'fake_value2']}
        mock_rpc_attachment_update.return_value = connection_info

        vref = tests_utils.create_volume(self.context, **volume_params)
        connector = {'fake': 'connector'}
        attachment = self.volume_api.attachment_create(self.context,
                                                       vref,
                                                       fake.UUID2,
                                                       connector)
        mock_rpc_attachment_update.assert_called_once_with(self.context,
                                                           mock.ANY,
                                                           connector,
                                                           mock.ANY)
        new_attachment = objects.VolumeAttachment.get_by_id(self.context,
                                                            attachment.id)
        self.assertEqual(connection_info, new_attachment.connection_info)

    @mock.patch.object(volume_api.API, 'attachment_deletion_allowed')
    @mock.patch('cinder.volume.rpcapi.VolumeAPI.attachment_delete')
    def test_attachment_delete_reserved(self,
                                        mock_rpc_attachment_delete,
                                        mock_allowed):
        """Test attachment_delete with reserved."""
        mock_allowed.return_value = None
        volume_params = {'status': 'available'}

        vref = tests_utils.create_volume(self.context, **volume_params)
        aref = self.volume_api.attachment_create(self.context,
                                                 vref,
                                                 fake.UUID2)
        aobj = objects.VolumeAttachment.get_by_id(self.context,
                                                  aref.id)
        self.assertEqual('reserved', aref.attach_status)
        self.assertEqual(vref.id, aref.volume_id)
        self.volume_api.attachment_delete(self.context,
                                          aobj)
        mock_allowed.assert_called_once_with(self.context, aobj)

        # Since it's just reserved and never finalized, we should never make an
        # rpc call
        mock_rpc_attachment_delete.assert_not_called()

    @mock.patch.object(volume_api.API, 'attachment_deletion_allowed')
    @mock.patch('cinder.volume.rpcapi.VolumeAPI.attachment_delete')
    @mock.patch('cinder.volume.rpcapi.VolumeAPI.attachment_update')
    def test_attachment_create_update_and_delete(
            self,
            mock_rpc_attachment_update,
            mock_rpc_attachment_delete,
            mock_allowed):
        """Test attachment_delete."""
        mock_allowed.return_value = None
        volume_params = {'status': 'available'}
        connection_info = {'fake_key': 'fake_value',
                           'fake_key2': ['fake_value1', 'fake_value2']}
        mock_rpc_attachment_update.return_value = connection_info

        vref = tests_utils.create_volume(self.context, **volume_params)
        aref = self.volume_api.attachment_create(self.context,
                                                 vref,
                                                 fake.UUID2)
        aref = objects.VolumeAttachment.get_by_id(self.context,
                                                  aref.id)
        vref = objects.Volume.get_by_id(self.context,
                                        vref.id)

        connector = {'fake': 'connector',
                     'host': 'somehost'}
        self.volume_api.attachment_update(self.context,
                                          aref,
                                          connector)
        aref = objects.VolumeAttachment.get_by_id(self.context,
                                                  aref.id)
        self.assertEqual(connection_info, aref.connection_info)
        # We mock the actual call that updates the status
        # so force it here
        values = {'volume_id': vref.id,
                  'volume_host': vref.host,
                  'attach_status': 'attached',
                  'instance_uuid': fake.UUID2}
        aref = db.volume_attach(self.context, values)

        aref = objects.VolumeAttachment.get_by_id(self.context,
                                                  aref.id)
        self.assertEqual(vref.id, aref.volume_id)
        self.volume_api.attachment_delete(self.context,
                                          aref)

        mock_allowed.assert_called_once_with(self.context, aref)
        mock_rpc_attachment_delete.assert_called_once_with(self.context,
                                                           aref.id,
                                                           mock.ANY)

    def test_additional_attachment_create_no_connector(self):
        """Test attachment_create no connector."""
        volume_params = {'status': 'available'}

        vref = tests_utils.create_volume(self.context, **volume_params)
        aref = self.volume_api.attachment_create(self.context,
                                                 vref,
                                                 fake.UUID2)
        self.assertEqual(fake.UUID2, aref.instance_uuid)
        self.assertIsNone(aref.attach_time)
        self.assertEqual('reserved', aref.attach_status)
        self.assertEqual('null', aref.attach_mode)
        self.assertEqual(vref.id, aref.volume_id)
        self.assertEqual({}, aref.connection_info)

        self.assertRaises(exception.InvalidVolume,
                          self.volume_api.attachment_create,
                          self.context,
                          vref,
                          fake.UUID1)
        self.volume_api.attachment_create(self.context,
                                          vref,
                                          fake.UUID2)
        vref = objects.Volume.get_by_id(self.context,
                                        vref.id)
        self.assertEqual(2, len(vref.volume_attachment))

    @mock.patch.object(volume_api.API, 'attachment_deletion_allowed')
    @mock.patch('cinder.volume.rpcapi.VolumeAPI.attachment_update')
    def test_attachment_create_reserve_delete(
            self,
            mock_rpc_attachment_update,
            mock_allowed):
        mock_allowed.return_value = None
        volume_params = {'status': 'available'}
        connector = {
            "initiator": "iqn.1993-08.org.debian:01:cad181614cec",
            "ip": "192.168.1.20",
            "platform": "x86_64",
            "host": "tempest-1",
            "os_type": "linux2",
            "multipath": False}

        connection_info = {'fake_key': 'fake_value',
                           'fake_key2': ['fake_value1', 'fake_value2']}
        mock_rpc_attachment_update.return_value = connection_info

        vref = tests_utils.create_volume(self.context, **volume_params)
        aref = self.volume_api.attachment_create(self.context,
                                                 vref,
                                                 fake.UUID2,
                                                 connector=connector)
        vref = objects.Volume.get_by_id(self.context,
                                        vref.id)
        # Need to set the status here because our mock isn't doing it for us
        vref.status = 'in-use'
        vref.save()

        # Now a second attachment acting as a reserve
        self.volume_api.attachment_create(self.context,
                                          vref,
                                          fake.UUID2)

        # We should now be able to delete the original attachment that gave us
        # 'in-use' status, and in turn we should revert to the outstanding
        # attachments reserve
        self.volume_api.attachment_delete(self.context,
                                          aref)
        mock_allowed.assert_called_once_with(self.context, aref)
        vref = objects.Volume.get_by_id(self.context,
                                        vref.id)
        self.assertEqual('reserved', vref.status)

    @mock.patch.object(volume_api.API, 'attachment_deletion_allowed')
    def test_reserve_reserve_delete(self, mock_allowed):
        """Test that we keep reserved status across multiple reserves."""
        mock_allowed.return_value = None
        volume_params = {'status': 'available'}

        vref = tests_utils.create_volume(self.context, **volume_params)
        aref = self.volume_api.attachment_create(self.context,
                                                 vref,
                                                 fake.UUID2)
        vref = objects.Volume.get_by_id(self.context,
                                        vref.id)
        self.assertEqual('reserved', vref.status)

        self.volume_api.attachment_create(self.context,
                                          vref,
                                          fake.UUID2)
        vref = objects.Volume.get_by_id(self.context,
                                        vref.id)
        self.assertEqual('reserved', vref.status)
        self.volume_api.attachment_delete(self.context,
                                          aref)
        mock_allowed.assert_called_once_with(self.context, aref)
        vref = objects.Volume.get_by_id(self.context,
                                        vref.id)
        self.assertEqual('reserved', vref.status)
        self.assertEqual(1, len(vref.volume_attachment))

    def test_attachment_create_readonly_volume(self):
        """Test attachment_create on a readonly volume."""
        volume_params = {'status': 'available'}

        vref = tests_utils.create_volume(self.context, **volume_params)
        self.volume_api.update_readonly_flag(self.context, vref, True)
        aref = self.volume_api.attachment_create(self.context,
                                                 vref,
                                                 fake.UUID2)
        self.assertEqual(fake.UUID2, aref.instance_uuid)
        self.assertIsNone(aref.attach_time)
        self.assertEqual('reserved', aref.attach_status)
        self.assertEqual('ro', aref.attach_mode)
        self.assertEqual(vref.id, aref.volume_id)
        self.assertEqual({}, aref.connection_info)

    def test_attachment_create_volume_in_error_state(self):
        """Test attachment_create volume in error state."""
        volume_params = {'status': 'available'}

        vref = tests_utils.create_volume(self.context, **volume_params)
        vref.status = "error"
        self.assertRaises(exception.InvalidVolume,
                          self.volume_api.attachment_create,
                          self.context,
                          vref,
                          fake.UUID2)

    def test_attachment_update_volume_in_error_state(self):
        """Test attachment_update volumem in error state."""
        volume_params = {'status': 'available'}

        vref = tests_utils.create_volume(self.context, **volume_params)
        aref = self.volume_api.attachment_create(self.context,
                                                 vref,
                                                 fake.UUID2)
        self.assertEqual(fake.UUID2, aref.instance_uuid)
        self.assertIsNone(aref.attach_time)
        self.assertEqual('reserved', aref.attach_status)
        self.assertEqual(vref.id, aref.volume_id)
        self.assertEqual({}, aref.connection_info)
        vref.status = 'error'
        vref.save()
        connector = {'fake': 'connector',
                     'host': 'somehost'}
        caught_exc = self.assertRaises(
            exception.ResourceConflict,
            self.volume_api.attachment_update,
            self.context, aref, connector)
        self.assertEqual(409, caught_exc.code)

    @mock.patch('cinder.db.sqlalchemy.api.volume_attachment_update',
                return_value={})
    @mock.patch('cinder.volume.rpcapi.VolumeAPI.attachment_update',
                return_value={})
    @mock.patch.object(db.sqlalchemy.api, '_volume_type_get',
                       v2_fakes.fake_volume_type_get)
    def test_attachment_update_duplicate(self, mock_va_update, mock_db_upd):
        volume_params = {'status': 'available'}

        vref = tests_utils.create_volume(self.context,
                                         deleted=0,
                                         **volume_params)

        tests_utils.attach_volume(self.context,
                                  vref.id,
                                  fake.UUID1,
                                  'somehost',
                                  'somemountpoint')

        # Update volume with another attachment
        tests_utils.attach_volume(self.context,
                                  vref.id,
                                  fake.UUID2,
                                  'somehost2',
                                  'somemountpoint2')
        vref.refresh()

        # This attachment will collide with the first
        connector = {'host': 'somehost'}
        vref.volume_attachment[0]['connector'] = {'host': 'somehost'}
        vref.volume_attachment[0]['connection_info'] = {'c': 'd'}
        with mock.patch('cinder.objects.Volume.get_by_id', return_value=vref):
            with mock.patch.object(self.volume_api.volume_rpcapi,
                                   'attachment_update') as m_au:
                caught_exc = self.assertRaises(
                    exception.ResourceConflict,
                    self.volume_api.attachment_update,
                    self.context,
                    vref.volume_attachment[1],
                    connector)
                self.assertEqual(409, caught_exc.code)
                m_au.assert_not_called()
        mock_va_update.assert_not_called()
        mock_db_upd.assert_not_called()

    def test_attachment_create_creating_volume(self):
        """Test attachment_create on a creating volume."""
        volume_params = {'status': 'creating'}

        vref = tests_utils.create_volume(self.context, **volume_params)
        self.assertRaises(exception.InvalidVolume,
                          self.volume_api.attachment_create,
                          self.context,
                          vref,
                          fake.UUID1)

    def _get_attachment(self, with_instance_id=True):
        volume = fake_volume.fake_volume_obj(self.context, id=fake.VOLUME_ID)
        volume.volume_attachment = objects.VolumeAttachmentList()
        attachment = fake_volume.volume_attachment_ovo(
            self.context,
            volume_id=fake.VOLUME_ID,
            instance_uuid=fake.INSTANCE_ID if with_instance_id else None,
            connection_info='{"a": 1}')
        attachment.volume = volume
        return attachment

    @mock.patch('cinder.compute.nova.API.get_server_volume')
    def test_attachment_deletion_allowed_service_call(self, mock_get_server):
        """Service calls are never redirected."""
        self.context.service_roles = ['reader', 'service']
        attachment = self._get_attachment()
        self.volume_api.attachment_deletion_allowed(self.context, attachment)
        mock_get_server.assert_not_called()

    @mock.patch('cinder.compute.nova.API.get_server_volume')
    def test_attachment_deletion_allowed_service_call_different_service_name(
            self, mock_get_server):
        """Service calls are never redirected and role can be different.

        In this test we support 2 different service roles, the standard service
        and a custom one called captain_awesome, and passing the custom one
        works as expected.
        """
        self.override_config('service_token_roles',
                             ['service', 'captain_awesome'],
                             group='keystone_authtoken')

        self.context.service_roles = ['reader', 'captain_awesome']
        attachment = self._get_attachment()
        self.volume_api.attachment_deletion_allowed(self.context, attachment)
        mock_get_server.assert_not_called()

    @mock.patch('cinder.compute.nova.API.get_server_volume')
    def test_attachment_deletion_allowed_no_instance(self, mock_get_server):
        """Attachments with no instance id are never redirected."""
        attachment = self._get_attachment(with_instance_id=False)
        self.volume_api.attachment_deletion_allowed(self.context, attachment)
        mock_get_server.assert_not_called()

    @mock.patch('cinder.compute.nova.API.get_server_volume')
    def test_attachment_deletion_allowed_no_conn_info(self, mock_get_server):
        """Attachments with no connection information are never redirected."""
        attachment = self._get_attachment(with_instance_id=False)
        attachment.connection_info = None
        self.volume_api.attachment_deletion_allowed(self.context, attachment)

        mock_get_server.assert_not_called()

    def test_attachment_deletion_allowed_no_attachment(self):
        """For users don't allow operation with no attachment reference."""
        self.assertRaises(exception.ConflictNovaUsingAttachment,
                          self.volume_api.attachment_deletion_allowed,
                          self.context, None)

    @mock.patch('cinder.objects.VolumeAttachment.get_by_id',
                side_effect=exception.VolumeAttachmentNotFound(filter=''))
    def test_attachment_deletion_allowed_attachment_id_not_found(self,
                                                                 mock_get):
        """For users don't allow if attachment cannot be found."""
        attachment = self._get_attachment(with_instance_id=False)
        attachment.connection_info = None
        self.assertRaises(exception.ConflictNovaUsingAttachment,
                          self.volume_api.attachment_deletion_allowed,
                          self.context, fake.ATTACHMENT_ID)
        mock_get.assert_called_once_with(self.context, fake.ATTACHMENT_ID)

    def test_attachment_deletion_allowed_volume_no_attachments(self):
        """For users allow if volume has no attachments."""
        volume = tests_utils.create_volume(self.context)
        self.volume_api.attachment_deletion_allowed(self.context, None, volume)

    def test_attachment_deletion_allowed_multiple_attachment(self):
        """For users don't allow if volume has multiple attachments."""
        attachment = self._get_attachment()
        volume = attachment.volume
        volume.volume_attachment = objects.VolumeAttachmentList(
            objects=[attachment, attachment])
        self.assertRaises(exception.ConflictNovaUsingAttachment,
                          self.volume_api.attachment_deletion_allowed,
                          self.context, None, volume)

    @mock.patch('cinder.compute.nova.API.get_server_volume')
    def test_attachment_deletion_allowed_vm_not_found(self, mock_get_server):
        """Don't reject if instance doesn't exist"""
        mock_get_server.side_effect = nova.API.NotFound(404)
        attachment = self._get_attachment()
        self.volume_api.attachment_deletion_allowed(self.context, attachment)

        mock_get_server.assert_called_once_with(self.context, fake.INSTANCE_ID,
                                                fake.VOLUME_ID)

    @mock.patch('cinder.compute.nova.API.get_server_volume')
    def test_attachment_deletion_allowed_attachment_from_volume(
            self, mock_get_server):
        """Don't reject if instance doesn't exist"""
        mock_get_server.side_effect = nova.API.NotFound(404)
        attachment = self._get_attachment()
        volume = attachment.volume
        volume.volume_attachment = objects.VolumeAttachmentList(
            objects=[attachment])
        self.volume_api.attachment_deletion_allowed(self.context, None, volume)

        mock_get_server.assert_called_once_with(self.context, fake.INSTANCE_ID,
                                                volume.id)

    @mock.patch('cinder.objects.VolumeAttachment.get_by_id')
    def test_attachment_deletion_allowed_mismatched_volume_and_attach_id(
            self, mock_get_attatchment):
        """Reject if volume and attachment don't match."""
        attachment = self._get_attachment()
        volume = attachment.volume
        volume.volume_attachment = objects.VolumeAttachmentList(
            objects=[attachment])
        attachment2 = self._get_attachment()
        attachment2.volume_id = attachment.volume.id = fake.VOLUME2_ID
        self.assertRaises(exception.InvalidInput,
                          self.volume_api.attachment_deletion_allowed,
                          self.context, attachment2.id, volume)
        mock_get_attatchment.assert_called_once_with(self.context,
                                                     attachment2.id)

    @mock.patch('cinder.objects.VolumeAttachment.get_by_id')
    @mock.patch('cinder.compute.nova.API.get_server_volume')
    def test_attachment_deletion_allowed_not_found_attachment_id(
            self, mock_get_server, mock_get_attachment):
        """Don't reject if instance doesn't exist"""
        mock_get_server.side_effect = nova.API.NotFound(404)
        mock_get_attachment.return_value = self._get_attachment()

        self.volume_api.attachment_deletion_allowed(self.context,
                                                    fake.ATTACHMENT_ID)

        mock_get_attachment.assert_called_once_with(self.context,
                                                    fake.ATTACHMENT_ID)

        mock_get_server.assert_called_once_with(self.context, fake.INSTANCE_ID,
                                                fake.VOLUME_ID)

    @mock.patch('cinder.compute.nova.API.get_server_volume')
    def test_attachment_deletion_allowed_mismatch_id(self, mock_get_server):
        """Don't reject if attachment id on nova doesn't match"""
        mock_get_server.return_value.attachment_id = fake.ATTACHMENT2_ID
        attachment = self._get_attachment()
        self.volume_api.attachment_deletion_allowed(self.context, attachment)

        mock_get_server.assert_called_once_with(self.context, fake.INSTANCE_ID,
                                                fake.VOLUME_ID)

    @mock.patch('cinder.compute.nova.API.get_server_volume')
    def test_attachment_deletion_allowed_user_call_fails(self,
                                                         mock_get_server):
        """Fail user calls"""
        attachment = self._get_attachment()
        mock_get_server.return_value.attachment_id = attachment.id
        self.assertRaises(exception.ConflictNovaUsingAttachment,
                          self.volume_api.attachment_deletion_allowed,
                          self.context, attachment)

        mock_get_server.assert_called_once_with(self.context, fake.INSTANCE_ID,
                                                fake.VOLUME_ID)



# Copyright 2016 Red Hat, Inc
# Copyright 2017 Rackspace Australia
# Copyright 2018 Michael Still and Aptira
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Helpers for cgroup related routines.
"""

import os.path

from oslo_concurrency import processutils

import cinder.privsep


@cinder.privsep.sys_admin_pctxt.entrypoint
def cgroup_create(name):
    # If this path exists, it means we have support for cgroups v2
    if os.path.isfile('/sys/fs/cgroup/cgroup.controllers'):
        # cgroups v2 doesn't support io, but blkio instead.
        processutils.execute('cgcreate', '-g', 'io:%s' % name)
    else:
        processutils.execute('cgcreate', '-g', 'blkio:%s' % name)


@cinder.privsep.sys_admin_pctxt.entrypoint
def cgroup_limit(name, rw, dev, bps):
    if os.path.isfile('/sys/fs/cgroup/cgroup.controllers'):
        if rw == 'read':
            cgset_arg = 'rbps'
        else:
            cgset_arg = 'wbps'
        processutils.execute('cgset', '-r',
                             'io.max=%s %s=%s' % (dev, cgset_arg, bps), name)
    else:
        processutils.execute('cgset', '-r',
                             'blkio.throttle.%s_bps_device=%s %d' % (rw, dev,
                                                                     bps),
                             name)
